empty_image_group_ontology = {
    "objects": [
        {"id": "1", "name": "Cars", "color": "#D33115", "shape": "bounding_box", "featureNodeHash": "r4M5dkj7"},
        {"id": "2", "name": "Horses", "color": "#E27300", "shape": "bounding_box", "featureNodeHash": "IuyN9D7S"},
        {"id": "3", "name": "Person", "color": "#16406C", "shape": "bounding_box", "featureNodeHash": "vZSR9ig7"},
        {"id": "4", "name": "Bin", "color": "#FE9200", "shape": "bounding_box", "featureNodeHash": "oYAJdhR1"},
        {"id": "5", "name": "Car segmentation", "color": "#FCDC00", "shape": "polygon", "featureNodeHash": "bPAym0bw"},
        {"id": "6", "name": "Bin", "color": "#DBDF00", "shape": "polygon", "featureNodeHash": "cFwx+HfR"},
        {"id": "7", "name": "Camel", "color": "#A4DD00", "shape": "bounding_box", "featureNodeHash": "oyEb2JCi"},
    ],
    "classifications": [
        {
            "id": "1",
            "featureNodeHash": "6nyvIs5W",
            "attributes": [
                {
                    "id": "1.1",
                    "name": "Is there a car in the frame? SUSPECTED",
                    "type": "radio",
                    "required": False,
                    "dynamic": False,
                    "featureNodeHash": "oehdT/Pe",
                    "options": [
                        {"id": "1.1.1", "label": "True", "value": "true", "featureNodeHash": "F/rpk4Fo"},
                        {"id": "1.1.2", "label": "False", "value": "false", "featureNodeHash": "/ILQSY6c"},
                        {"id": "1.1.3", "label": "Maybe", "value": "maybe", "featureNodeHash": "EDWbo8P8"},
                        {"id": "1.1.4", "label": "Maybe not", "value": "maybe_not", "featureNodeHash": "Uxzj/Efq"},
                        {
                            "id": "1.1.5",
                            "label": "SUSPECTED this is not the case",
                            "value": "suspected_this_is_not_the_case",
                            "featureNodeHash": "ZrRZBwx3",
                        },
                    ],
                }
            ],
        },
        {
            "id": "2",
            "featureNodeHash": "R0SMUli5",
            "attributes": [
                {
                    "id": "2.1",
                    "name": "This is a test",
                    "type": "radio",
                    "required": False,
                    "featureNodeHash": "vDUhRPP0",
                    "options": [
                        {"id": "2.1.1", "label": "Test", "value": "test", "featureNodeHash": "P7BwGqM3"},
                        {"id": "2.1.2", "label": "TEst2", "value": "test2", "featureNodeHash": "+LDYFQp1"},
                    ],
                }
            ],
        },
    ],
}


empty_image_group_labels = {
    "label_hash": "f1e6bc82-9f89-4545-8abb-f271bf28cf99",
    "branch_name": "main",
    "created_at": "Thu, 09 Feb 2023 14:12:03 UTC",
    "last_edited_at": "Thu, 09 Feb 2023 14:12:03 UTC",
    "data_hash": "aaa6bc82-9f89-4545-adbb-f271bf28cf99",
    "annotation_task_status": "QUEUED",
    "is_shadow_data": False,
    "dataset_hash": "d9f19c3c-5cd0-4f8c-b98c-6c0e24676224",
    "dataset_title": "One image group",
    "data_title": "image-group-8375e",
    "data_type": "img_group",
    "data_units": {
        "f850dfb4-7146-49e0-9afc-2b9434a64a9f": {
            "data_hash": "f850dfb4-7146-49e0-9afc-2b9434a64a9f",
            "data_title": "Screenshot 2021-11-24 at 18.35.57.png",
            "data_link": "https://storage.googleapis.com/cord-ai-platform.appspot.com/cord-images-prod/yiA5JxmLEGSoEcJAuxr3AJdDDXE2/f850dfb4-7146-49e0-9afc-2b9434a64a9f?X-Goog-Algorithm=GOOG4-RSA-SHA256&X-Goog-Credential=firebase-adminsdk-64w1p%40cord-ai-platform.iam.gserviceaccount.com%2F20221201%2Fauto%2Fstorage%2Fgoog4_request&X-Goog-Date=20221201T133838Z&X-Goog-Expires=604800&X-Goog-SignedHeaders=host&X-Goog-Signature=94c66d85014ff52a99fec1cf671ccc1b859ebead4308ca82c4d810e13ac285d2afa8cfa4bfcbd09f615b243b95d9b1d5d1d779e7a4ba5832a2207b4f3b99dbe405ded373f03f06abe4e24098e70568c269899f2f397c7a4392a1c3090bff2b8c98f2177f5db36f0884a83033f404354bdfda0506bf162e25ff6186fc54104e8273e86959b0296958a03359514660528a54ba94e25c59e59534ce5102f9c87ff7cb03a591606b3a191123af4a30fa4296a788a9433f0c8c1dc7d3f80a022cc42f8716ba44d09ecd04118dc6e4ee5977ffbadcc8d635cc4e906f024dba26e520cfc304fc0f3458a3e3b2422c196956fd3024a6eba0512d557683487b10a1a381b4",
            "data_type": "image/png",
            "data_sequence": "0",
            "width": 952,
            "height": 678,
            "labels": {
                "objects": [],
                "classifications": [],
            },
        },
        "177d1bb7-5394-4772-ba9f-4569f0c2a995": {
            "data_hash": "177d1bb7-5394-4772-ba9f-4569f0c2a995",
            "data_title": "Screenshot 2021-11-24 at 18.36.02.png",
            "data_link": "https://storage.googleapis.com/cord-ai-platform.appspot.com/cord-images-prod/yiA5JxmLEGSoEcJAuxr3AJdDDXE2/177d1bb7-5394-4772-ba9f-4569f0c2a995?X-Goog-Algorithm=GOOG4-RSA-SHA256&X-Goog-Credential=firebase-adminsdk-64w1p%40cord-ai-platform.iam.gserviceaccount.com%2F20221201%2Fauto%2Fstorage%2Fgoog4_request&X-Goog-Date=20221201T133838Z&X-Goog-Expires=604800&X-Goog-SignedHeaders=host&X-Goog-Signature=808ea15115717eaa939b6b66a74c2d8771216a16d621d7ff8a002e5bec65fcc3053097e2bba0711e3442b4a41afe9744e42ffe6b08194cd83debcf3dbb2d649ab2a92df2b02ed4e77e8eedef577007bb33681410ec164e24ba4c5643df744687762069e9dd360c0fa0820e90cd977d0c199b391e1425dd0e696a517e747b70ed4953f514be87846317c762c519af1f49066797d7a5f9e476bf9abf06e62a3ed5ebcf05516782e2ec913af68a6bedac0349c6b2a528e5411ec96275326c0e67d699cea81eadb6660e31d4a7d611db64b06f67098bad266c7bc6eb6e5191b7a97101a01ffd61b40aaaf24dea2026add406ec362df4df8eac26750b569295c28b86",
            "data_type": "image/png",
            "data_sequence": "1",
            "width": 952,
            "height": 678,
            "labels": {"objects": [], "classifications": []},
        },
        "5f393aff-38e3-4364-8ece-340d5aade1c4": {
            "data_hash": "5f393aff-38e3-4364-8ece-340d5aade1c4",
            "data_title": "Screenshot 2021-11-24 at 18.36.07.png",
            "data_link": "https://storage.googleapis.com/cord-ai-platform.appspot.com/cord-images-prod/yiA5JxmLEGSoEcJAuxr3AJdDDXE2/5f393aff-38e3-4364-8ece-340d5aade1c4?X-Goog-Algorithm=GOOG4-RSA-SHA256&X-Goog-Credential=firebase-adminsdk-64w1p%40cord-ai-platform.iam.gserviceaccount.com%2F20221201%2Fauto%2Fstorage%2Fgoog4_request&X-Goog-Date=20221201T133838Z&X-Goog-Expires=604800&X-Goog-SignedHeaders=host&X-Goog-Signature=8929d2c90b42f74e57fd7b7e2a8e8d9e531d030808c279a2ebb30ddfc589ae4524d28c611fb67b2dfb77a329c2cf48ec21bcc9cbb0d24ac3c822f7c6b47c9f02bf33b45e9e810a500b289d61cd2dec8ea24b8b2d0f5287db8325288f51ac049fd6c3a28c85b29c5ac08634302e7d4c6afaec779d9e11148220f5412532ff152c36055a5fd197c8bdd3fe5471c36c061cf52c91607a5bb9a07a17aa352a59a170a3be7e4010123622d7a82b5714cc9b2a47bc320938973911e760e2b6d919ccfdf3da6cafdfedb888989ad1306581f30c287259fdf1a29c69b34b6ec4a91d332356c918281bac45ef8ff3665a3b1b0990b225d588b7897097a06136a5dd1bbbc7",
            "data_type": "image/png",
            "data_sequence": "2",
            "width": 952,
            "height": 678,
            "labels": {"objects": [], "classifications": []},
        },
        "6518e742-8dee-4a6d-b491-710f8a5c0fc4": {
            "data_hash": "6518e742-8dee-4a6d-b491-710f8a5c0fc4",
            "data_title": "Screenshot 2021-11-24 at 18.36.11.png",
            "data_link": "https://storage.googleapis.com/cord-ai-platform.appspot.com/cord-images-prod/yiA5JxmLEGSoEcJAuxr3AJdDDXE2/6518e742-8dee-4a6d-b491-710f8a5c0fc4?X-Goog-Algorithm=GOOG4-RSA-SHA256&X-Goog-Credential=firebase-adminsdk-64w1p%40cord-ai-platform.iam.gserviceaccount.com%2F20221201%2Fauto%2Fstorage%2Fgoog4_request&X-Goog-Date=20221201T133838Z&X-Goog-Expires=604800&X-Goog-SignedHeaders=host&X-Goog-Signature=001d6367d2bba32a7483a5217e06a5b1855fe0ddc18d0a2e33482afec25f5f2841f352e115c8d63d38f306b2514e19e021436e0bcc83177a85802d28fbe08f6bd6021331c5ca5cce3f7383a9190f56748378880a1b35d4ab0913716727c10214a11989fbcc6e3b52b0b4df382562c93ffb175ba3e87ce4a1494c5739b716a5ff0b4570bd69074642f0960f1ecfa333312accb1283016ba427749fa2129ac36927304f0fd5b722850084521b874fd53e63fd7dab25fcd20e7caf7fa3021f13b8b7ef0747c36848f6fe3f727937ed6f0243ad1e69d3d46213b6735c349f498534d7c55a106ae61042128fe7d80e4773897f28957d346844bea82b845e4856309ff",
            "data_type": "image/png",
            "data_sequence": "3",
            "width": 952,
            "height": 678,
            "labels": {"objects": [], "classifications": []},
        },
        "24dd04ee-ff3f-4f91-b6b5-b385defc7307": {
            "data_hash": "24dd04ee-ff3f-4f91-b6b5-b385defc7307",
            "data_title": "Screenshot 2021-11-24 at 18.36.19.png",
            "data_link": "https://storage.googleapis.com/cord-ai-platform.appspot.com/cord-images-prod/yiA5JxmLEGSoEcJAuxr3AJdDDXE2/24dd04ee-ff3f-4f91-b6b5-b385defc7307?X-Goog-Algorithm=GOOG4-RSA-SHA256&X-Goog-Credential=firebase-adminsdk-64w1p%40cord-ai-platform.iam.gserviceaccount.com%2F20221201%2Fauto%2Fstorage%2Fgoog4_request&X-Goog-Date=20221201T133838Z&X-Goog-Expires=604800&X-Goog-SignedHeaders=host&X-Goog-Signature=4574efaee32c7f3aa37fb5d5e47718e79685b802c9563861bfc057c25c8e3d85672fe04daaa694b39d3ebc6f60589c68e8ccf810d05f896896136cf178eb5a24ebb67b37501ce78c37b3618b1cde1a515ff8e07d9c6d9afb6476d4d792845ea954251c3383dd56b0d76d62be55644df2eeb36c9db06fb01e905a8659b1e8170344dcd1c65ac001e5e3e38ae49cd0aa5780681901177b0b0743d67b3b28ac698590e53a7b3ae2da6f9bbb1e2c15600a1d5a249b7fddbb7ee92f7a0684a71d6b7919f5cb2f9171e1dd47afb04775f3153b265f78f3b74355057c68c7e40ca12b50880925029d51e014c5a15bd1e8d40921f69dca5764e2bedc472c6915789b8b8b",
            "data_type": "image/png",
            "data_sequence": "4",
            "width": 952,
            "height": 678,
            "labels": {"objects": [], "classifications": []},
        },
    },
    "object_answers": {},
    "classification_answers": {},
    "object_actions": {},
    "label_status": "LABEL_IN_PROGRESS",
}



"""
Code Block Name: Frmae range for a specific task
"""

# Import dependencies
import json

from encord import EncordUserClient
from encord.objects import ObjectInstance
from encord.objects.attributes import Attribute

# User input
SSH_PATH = "/Users/chris-encord/ssh-private-key.txt"
PROJECT_ID = "00000000-0000-0000-0000-000000000000"
DATA_UNIT = "cherries-is"
OUTPUT_FILE_PATH = "/Users/chris-encord/frame_range_output.json"
START_FRAME_NUMBER = 0
END_FRAME_NUMBER = 35
BUNDLE_SIZE = 100

# Create user client using SSH key
user_client: EncordUserClient = EncordUserClient.create_with_ssh_private_key(
    ssh_private_key_path=SSH_PATH,
    # For US platform users use "https://api.us.encord.com"
    domain="https://api.encord.com",
)

# Load the project
project = user_client.get_project(PROJECT_ID)
assert project is not None, f"Project with ID {PROJECT_ID} could not be loaded"

# Get filtered label rows
label_rows = project.list_label_rows_v2(data_title_eq=DATA_UNIT)
assert label_rows, f"No label rows found for data unit: {DATA_UNIT}"

# Initialize labels using bundles
with project.create_bundle(bundle_size=BUNDLE_SIZE) as bundle:
    for label_row in label_rows:
        label_row.initialise_labels(bundle=bundle)


# Function to extract attributes in a structured format
def extract_attributes(attribute: Attribute, object_instance: ObjectInstance):
    try:
        answer = object_instance.get_answer(attribute)
    except Exception:
        answer = None
    return {
        "attribute_name": attribute.title,
        "attribute_hash": attribute.feature_node_hash,
        "attribute_answer": str(answer),
    }


# Result collection
results = []

# Loop over label rows
for label_row in label_rows:
    row_data = {
        "label_row_hash": label_row.label_hash,
        "data_title": label_row.data_title,
        "objects": [],
        "classifications": [],
    }

    # Process object annotations
    object_instances = label_row.get_object_instances()
    assert object_instances, f"No object instances found in label row {label_row.uid}"

    for object_instance in object_instances:
        annotations = object_instance.get_annotations()
        assert annotations, f"No annotations found for object instance {object_instance.object_hash}"

        assert object_instance.ontology_item and object_instance.ontology_item.attributes, (
            f"No attributes found for object {object_instance.object_hash}"
        )

        for annotation in annotations:
            if START_FRAME_NUMBER <= annotation.frame <= END_FRAME_NUMBER:
                obj_info = {
                    "frame": annotation.frame,
                    "object_hash": object_instance.object_hash,
                    "object_name": object_instance.object_name,
                    "feature_hash": object_instance.feature_hash,
                    "uid": str(object_instance.ontology_item.uid),
                    "color": object_instance.ontology_item.color,
                    "shape": object_instance.ontology_item.shape,
                    "label_location": str(annotation.coordinates),
                    "attributes": [],
                }

                print("Frame:", annotation.frame)
                print("objectHash:", object_instance.object_hash)
                print("Object name:", object_instance.object_name)
                print("featureHash:", object_instance.feature_hash)
                print("uid:", obj_info["uid"])
                print("Object color:", object_instance.ontology_item.color)
                print("Ontology shape:", object_instance.ontology_item.shape)
                print(f"Label location: {annotation.coordinates}")

                for attribute in object_instance.ontology_item.attributes:
                    attr_data = extract_attributes(attribute, object_instance)
                    obj_info["attributes"].append(attr_data)

                    print(f"Attribute name: {attr_data['attribute_name']}")
                    print(f"Attribute hash: {attr_data['attribute_hash']}")
                    print(f"Attribute answer: {attr_data['attribute_answer']}")

                row_data["objects"].append(obj_info)

    # Process classification annotations
    classification_instances = label_row.get_classification_instances()
    assert classification_instances is not None, f"No classification instances found in label row {label_row.uid}"

    for classification_instance in classification_instances:
        annotations = classification_instance.get_annotations()
        assert annotations, (
            f"No annotations found for classification instance {classification_instance.classification_hash}"
        )

        for annotation in annotations:
            if START_FRAME_NUMBER <= annotation.frame <= END_FRAME_NUMBER:
                try:
                    answer = classification_instance.get_answer()
                    answer_value = answer.value
                    answer_hash = answer.feature_node_hash
                except Exception:
                    answer_value = None
                    answer_hash = None

                classification_info = {
                    "frame": annotation.frame,
                    "classification_hash": classification_instance.classification_hash,
                    "classification_name": classification_instance.classification_name,
                    "feature_hash": classification_instance.feature_hash,
                    "value": answer_value,
                    "answer_hash": answer_hash,
                }

                print("Classification hash:", classification_info["classification_hash"])
                print("Classification name:", classification_info["classification_name"])
                print("Feature hash:", classification_info["feature_hash"])
                print("Classification value:", classification_info["value"])
                print("Classification answer hash:", classification_info["answer_hash"])

                row_data["classifications"].append(classification_info)

    results.append(row_data)

# Save to JSON
assert OUTPUT_FILE_PATH.endswith(".json"), "Output file path must be a JSON file"

with open(OUTPUT_FILE_PATH, "w") as f:
    json.dump(results, f, indent=4)

print(f"\nData saved to: {OUTPUT_FILE_PATH}")



"""---
title: "Objects - Frames"
slug: "sdk-ref-objects-frames"
hidden: false
metadata:
  title: "Objects - Frames"
  description: "Encord SDK Objects - Frames."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Collection, List, Union, cast


@dataclass
class Range:
    """A class representing a range with a start and end value.

    Attributes:
        start (int): The starting value of the range.
        end (int): The ending value of the range.
    """

    start: int
    end: int

    def __repr__(self):
        return f"({self.start}:{self.end})"

    def overlaps(self, other) -> bool:
        """Check if two ranges overlap"""
        return self.start <= other.end and other.start <= self.end

    def merge(self, other) -> None:
        """Merge another range into this one."""
        self.start = min(self.start, other.start)
        self.end = max(self.end, other.end)

    def copy(self) -> Range:
        return Range(self.start, self.end)


Ranges = List[Range]
FramesList = List[int]
Frames = Union[int, FramesList, Range, Ranges]


def frame_to_range(frame: int) -> Range:
    """Convert a single frame to a Range.

    Args:
        frame (int): The single frame to be converted.

    Returns:
        Range: A Range object with both start and end set to the input frame.
    """
    return Range(frame, frame)


def frames_to_ranges(frames: Collection[int]) -> Ranges:
    """Create a sorted list (in ascending order) of non-overlapping run-length encoded ranges from a collection of frames.

    Args:
        frames (Collection[int]): A collection of integers representing frames.

    Returns:
        Ranges: A list of Range objects representing the non-overlapping ranges.
    """
    if len(frames) == 0:
        return []

    ret = []

    frames_sorted = sorted(frames)
    last_value = frames_sorted[0]
    next_range = Range(start=last_value, end=last_value)
    idx = 1
    while idx < len(frames_sorted):
        if frames_sorted[idx] == last_value + 1:
            next_range.end = frames_sorted[idx]
        else:
            ret.append(next_range)
            next_range = Range(frames_sorted[idx], frames_sorted[idx])
        last_value = frames_sorted[idx]
        idx += 1

    ret.append(next_range)
    return ret


def ranges_to_list(ranges: Ranges) -> List[List[int]]:
    """Convert a list of Range objects to a list of lists (run-length encoded) of integers.

    Args:
        ranges (Ranges): A list of Range objects.

    Returns:
        List[List[int]]: A list of lists where each inner list contains two integers, the start and end of a range.
    """
    return [[r.start, r.end] for r in ranges]


def range_to_ranges(range_: Range) -> Ranges:
    """Convert a single Range to a list of Ranges.

    Args:
        range_ (Range): The single Range to be converted.

    Returns:
        Ranges: A list containing the input Range as its only element.
    """
    return [range_]


def range_to_frames(range_: Range) -> List[int]:
    """Convert a single Range (run-length encoded) to a list of integers.

    Args:
        range_ (Range): The single Range to be converted.

    Returns:
        List[int]: A list of integers representing the frames within the range.
    """
    return list(range(range_.start, range_.end + 1))


def ranges_to_frames(range_list: Ranges) -> List[int]:
    """Convert a list of Ranges (run-length encoded) to a list of integers.

    Args:
        range_list (Ranges): A list of Range objects.

    Returns:
        List[int]: A sorted list of integers representing all frames within the ranges.
    """
    frames = set()
    for range_ in range_list:
        frames |= set(range_to_frames(range_))
    return sorted(list(frames))


def ranges_list_to_ranges(range_list: List[List[int]]) -> Ranges:
    """Convert a list of lists (run-length encoded) of integers to a list of Range objects.

    Args:
        range_list (List[List[int]]): A list of lists where each inner list contains two integers, the start and end of a range.

    Returns:
        Ranges: A list of Range objects created from the input list of lists.
    """
    return [Range(start, end) for start, end in range_list]


def frames_class_to_frames_list(frames_class: Frames) -> List[int]:
    """Convert a Frames class (which can be an int, a list of ints, a Range, or a list of Ranges) to a list of integers.

    Args:
        frames_class (Frames): A Frames class which can be a single int, a list of ints, a Range object, or a list of Range objects.

    Returns:
        List[int]: A sorted list of integers representing all frames within the input Frames class.

    Raises:
        RuntimeError: If the input frames_class is of an unexpected type.
    """
    if isinstance(frames_class, int):
        return [frames_class]
    elif isinstance(frames_class, Range):
        return range_to_frames(frames_class)
    elif isinstance(frames_class, list):
        if all(isinstance(x, int) for x in frames_class):
            return sorted(list(set(cast(List[int], frames_class))))
        elif all(isinstance(x, Range) for x in frames_class):
            return ranges_to_frames(cast(List[Range], frames_class))

    raise RuntimeError("Unexpected type for frames.")



"""
Code Block Name: Add Datasets to Project
"""

from encord import EncordUserClient, Project

# User input
SSH_PATH = "/Users/chris-encord/ssh-private-key.txt"
PROJECT_ID = "00000000-0000-0000-0000-000000000000"
DATASET_ID_01 = "00000000-0000-0000-0000-000000000000"
DATASET_ID_02 = "00000000-0000-0000-0000-000000000000"
DATASET_ID_03 = "00000000-0000-0000-0000-000000000000"

# Create user client using SSH key
user_client: EncordUserClient = EncordUserClient.create_with_ssh_private_key(
    ssh_private_key_path=SSH_PATH,
    # For US platform users use "https://api.us.encord.com"
    domain="https://api.encord.com",
)

# Open the project you want to work on by specifying the Project ID
project = user_client.get_project(PROJECT_ID)

add_these_datasets = project.add_datasets(
    [
        DATASET_ID_01,
        DATASET_ID_02,
        DATASET_ID_03,
        # ...
    ]
)
print(f"All Datasets for Project= {project.list_datasets()}")



"""
Code Block Name: All labels all branches
"""

import json

from encord import EncordUserClient

# User input
SSH_PATH = "/Users/chris-encord/ssh-private-key.txt"
PROJECT_ID = "00000000-0000-0000-0000-000000000000"
OUTPUT_FILE_PATH = "/Users/chris-encord/frame_range_output.json"
BUNDLE_SIZE = 100

# Create user client using SSH key
user_client: EncordUserClient = EncordUserClient.create_with_ssh_private_key(
    ssh_private_key_path=SSH_PATH,
    # For US platform users use "https://api.us.encord.com"
    domain="https://api.encord.com",
)

# Specify Project. Replace <project_id> with the ID of the Project you want to export labels for.
project = user_client.get_project(PROJECT_ID)

# Downloads a local copy of all the labels
# Without the include_all_label_branches flag only the MAIN branch labels export
label_rows = project.list_label_rows_v2(include_all_label_branches=True)

output_data = []  # This will hold the output data to be saved as JSON

# Initialize label rows using bundles
with project.create_bundle(bundle_size=BUNDLE_SIZE) as bundle:
    for label_row in label_rows:
        label_row.initialise_labels(bundle=bundle)

# Collecting data for JSON output
for label_row in label_rows:
    label_row_data = {
        "title": label_row.data_title,
        "branch": label_row.branch_name,
        "objects": [],
        "classifications": [],
    }

    # Collect object instances
    for object_instance in label_row.get_object_instances():
        object_data = {
            "object_hash": object_instance.object_hash,
            "object_name": object_instance.object_name,
            "feature_hash": object_instance.feature_hash,
            "ontology_item": {
                "uid": object_instance.ontology_item.uid,
                "color": object_instance.ontology_item.color,
                "shape": object_instance.ontology_item.shape,
                "attributes": [
                    {"name": attribute.name, "value": attribute.value}
                    for attribute in object_instance.ontology_item.attributes
                ],
            },
            "annotations": [
                {"frame": annotation.frame, "coordinates": annotation.coordinates}
                for annotation in object_instance.get_annotations()
            ],
        }
        label_row_data["objects"].append(object_data)

    # Collect classification instances
    for classification_instance in label_row.get_classification_instances():
        classification_data = {
            "classification_hash": classification_instance.classification_hash,
            "classification_name": classification_instance.classification_name,
            "feature_hash": classification_instance.feature_hash,
            "classification_answer": {
                "value": classification_instance.get_answer().value,
                "hash": classification_instance.get_answer().feature_node_hash,
            },
            "annotations": [{"frame": annotation.frame} for annotation in classification_instance.get_annotations()],
        }
        label_row_data["classifications"].append(classification_data)

    # Add label row data to the output list
    output_data.append(label_row_data)

# Saving to JSON file
output_file_path = OUTPUT_FILE_PATH  # Replace with the desired file path
with open(output_file_path, "w") as json_file:
    json.dump(output_data, json_file, indent=4)

print(f"Output saved to {output_file_path}")



"""
Code Block Name: Rotatable Bounding Boxes Images/Videos
"""

# Import dependencies
from encord import EncordUserClient, Project
from encord.objects import ChecklistAttribute, Object, ObjectInstance, Option, RadioAttribute, TextAttribute
from encord.objects.coordinates import RotatableBoundingBoxCoordinates

# User input
SSH_PATH = "/Users/chris-encord/ssh-private-key.txt"
PROJECT_ID = "00000000-0000-0000-0000-000000000000"
BUNDLE_SIZE = 100

# Create user client using ssh key
user_client: EncordUserClient = EncordUserClient.create_with_ssh_private_key(
    ssh_private_key_path=SSH_PATH,
    # For US platform users use "https://api.us.encord.com"
    domain="https://api.encord.com",
)

# Get project for which labels are to be added
project: Project = user_client.get_project(PROJECT_ID)
assert project is not None, "Project not found — check PROJECT_ID"

# Create radio button attribute for persimmon type
ontology_structure = project.ontology_structure
assert ontology_structure is not None, "Ontology structure is missing from the project"

# Find a bounding box annotation object in the project ontology
rbbox_ontology_object: Object = ontology_structure.get_child_by_title(title="Persimmons", type_=Object)
assert rbbox_ontology_object is not None, "Object 'Persimmons' not found in ontology"

# Get radio attribute for persimmon type
persimmon_type_radio_attribute = ontology_structure.get_child_by_title(type_=RadioAttribute, title="Type?")
assert persimmon_type_radio_attribute is not None, "Radio attribute 'Type?' not found"

# Create radio options
hachiya_option = persimmon_type_radio_attribute.get_child_by_title(type_=Option, title="Hachiya")
assert hachiya_option is not None, "Option 'Hachiya' not found under 'Type?'"

fuyu_option = persimmon_type_radio_attribute.get_child_by_title(type_=Option, title="Fuyu")
assert fuyu_option is not None, "Option 'Fuyu' not found under 'Type?'"

rjb_option = persimmon_type_radio_attribute.get_child_by_title(type_=Option, title="Rojo Brilliante")
assert rjb_option is not None, "Option 'Rojo Brilliante' not found under 'Type?'"

other_persimmon_option = persimmon_type_radio_attribute.get_child_by_title(type_=Option, title="Other persimmon type")
assert other_persimmon_option is not None, "Option 'Other persimmon type' not found under 'Type?'"

# Hachiya Qualities
hachiya_checklist_attribute = ontology_structure.get_child_by_title(
    type_=ChecklistAttribute, title="Hachiya Qualities?"
)
assert hachiya_checklist_attribute is not None, "Checklist attribute 'Hachiya Qualities?' not found"

hachiya_plump_option = hachiya_checklist_attribute.get_child_by_title(type_=Option, title="Plump")
assert hachiya_plump_option is not None, "Option 'Plump' not found under 'Hachiya Qualities?'"

hachiya_juicy_option = hachiya_checklist_attribute.get_child_by_title(type_=Option, title="Juicy")
assert hachiya_juicy_option is not None, "Option 'Juicy' not found under 'Hachiya Qualities?'"

hachiya_large_option = hachiya_checklist_attribute.get_child_by_title(type_=Option, title="Large")
assert hachiya_large_option is not None, "Option 'Large' not found under 'Hachiya Qualities?'"

# Fuyu Qualities
fuyu_checklist_attribute = ontology_structure.get_child_by_title(type_=ChecklistAttribute, title="Fuyu Qualities?")
assert fuyu_checklist_attribute is not None, "Checklist attribute 'Fuyu Qualities?' not found"

fuyu_plump_option = fuyu_checklist_attribute.get_child_by_title(type_=Option, title="Plump")
assert fuyu_plump_option is not None, "Option 'Plump' not found under 'Fuyu Qualities?'"

fuyu_juicy_option = fuyu_checklist_attribute.get_child_by_title(type_=Option, title="Juicy")
assert fuyu_juicy_option is not None, "Option 'Juicy' not found under 'Fuyu Qualities?'"

fuyu_large_option = fuyu_checklist_attribute.get_child_by_title(type_=Option, title="Large")
assert fuyu_large_option is not None, "Option 'Large' not found under 'Fuyu Qualities?'"

# Rojo Brilliante Qualities
rjb_checklist_attribute = ontology_structure.get_child_by_title(
    type_=ChecklistAttribute, title="Rojo Brilliante Qualities?"
)
assert rjb_checklist_attribute is not None, "Checklist attribute 'Rojo Brilliante Qualities?' not found"

rjb_plump_option = rjb_checklist_attribute.get_child_by_title(type_=Option, title="Plump")
assert rjb_plump_option is not None, "Option 'Plump' not found under 'Rojo Brilliante Qualities?'"

rjb_juicy_option = rjb_checklist_attribute.get_child_by_title(type_=Option, title="Juicy")
assert rjb_juicy_option is not None, "Option 'Juicy' not found under 'Rojo Brilliante Qualities?'"

rjb_large_option = rjb_checklist_attribute.get_child_by_title(type_=Option, title="Large")
assert rjb_large_option is not None, "Option 'Large' not found under 'Rojo Brilliante Qualities?'"

# Other persimmon type text input
other_persimmon_option_text_attribute = ontology_structure.get_child_by_title(
    type_=TextAttribute, title="Specify persimmon type"
)
assert other_persimmon_option_text_attribute is not None, "Text attribute 'Specify persimmon type' not found"

# Dictionary of labels per data unit and per frame with persimmon type specified, including quality options
video_frame_labels = {
    "cherries-001.jpg": {
        0: {
            "label_ref": "persimmon_001",
            "coordinates": RotatableBoundingBoxCoordinates(
                height=0.4, width=0.4, top_left_x=0.05, top_left_y=0.55, theta=23
            ),
            "persimmon_type": "Rojo Brilliante",
            "rjb_quality_options": "Plump, Juicy",
        },
    },
    "cherries-010.jpg": {
        0: [
            {
                "label_ref": "persimmon_002",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.4, width=0.4, top_left_x=0.05, top_left_y=0.55, theta=23
                ),
                "persimmon_type": "Fuyu",
                "fuyu_quality_options": "Plump, Juicy, Large",
            },
            {
                "label_ref": "persimmon_003",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.2, width=0.1, top_left_x=0.2, top_left_y=0.2, theta=25
                ),
                "persimmon_type": "Rojo Brilliante",
                "rjb_quality_options": "Plump",
            },
            {
                "label_ref": "persimmon_004",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.2, width=0.1, top_left_x=0.3, top_left_y=0.3, theta=27
                ),
                "persimmon_type": "Other persimmon type",
                "Specify persimmon type": "Triumph",
            },
        ],
    },
    "cherries-ig": {
        0: {
            "label_ref": "persimmon_005",
            "coordinates": RotatableBoundingBoxCoordinates(
                height=0.4, width=0.4, top_left_x=0.05, top_left_y=0.55, theta=23
            ),
            "persimmon_type": "Hachiya",
            "hachiya_quality_options": "Plump, Juicy",
        },
        2: [
            {
                "label_ref": "persimmon_006",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.1, width=0.2, top_left_x=0.2, top_left_y=0.2, theta=23
                ),
                "persimmon_type": "Fuyu",
                "fuyu_quality_options": "Large",
            },
            {
                "label_ref": "persimmon_007",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.1, width=0.2, top_left_x=0.4, top_left_y=0.2, theta=127
                ),
                "persimmon_type": "Rojo Brilliante",
                "rjb_quality_options": "Plump",
            },
            {
                "label_ref": "persimmon_008",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.2, width=0.1, top_left_x=0.6, top_left_y=0.3, theta=137
                ),
                "persimmon_type": "Other persimmon type",
                "Specify persimmon type": "Hiro",
            },
        ],
    },
    "cherries-is": {
        0: {
            "label_ref": "persimmon_009",
            "coordinates": RotatableBoundingBoxCoordinates(
                height=0.4, width=0.4, top_left_x=0.05, top_left_y=0.55, theta=23
            ),
            "persimmon_type": "Hachiya",
            "hachiya_quality_options": "Plump",
        },
        3: [
            {
                "label_ref": "persimmon_010",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.4, width=0.4, top_left_x=0.05, top_left_y=0.05, theta=23
                ),
                "persimmon_type": "Fuyu",
                "fuyu_quality_options": "Plump, Juicy, Large",
            },
            {
                "label_ref": "persimmon_011",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.3, width=0.3, top_left_x=0.5, top_left_y=0.05, theta=23
                ),
                "persimmon_type": "Rojo Brilliante",
                "rjb_quality_options": "Plump, Juicy",
            },
            {
                "label_ref": "persimmon_012",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.2, width=0.1, top_left_x=0.7, top_left_y=0.05, theta=23
                ),
                "persimmon_type": "Other persimmon type",
                "Specify persimmon type": "Maru",
            },
        ],
    },
    "cherries-vid-001.mp4": {
        103: [
            {
                "label_ref": "persimmon_013",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.5, width=0.5, top_left_x=0.05, top_left_y=0.55, theta=23
                ),
                "persimmon_type": "Rojo Brilliante",
                "rjb_quality_options": "Plump",
            },
            {
                "label_ref": "persimmon_014",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.2, width=0.2, top_left_x=0.3, top_left_y=0.2, theta=23
                ),
                "persimmon_type": "Hachiya",
                "hachiya_quality_options": "Plump, Juicy, Large",
            },
            {
                "label_ref": "persimmon_015",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.2, width=0.1, top_left_x=0.6, top_left_y=0.2, theta=23
                ),
                "persimmon_type": "Other persimmon type",
                "Specify persimmon type": "Triumph",
            },
        ],
        104: [
            {
                "label_ref": "persimmon_016",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.5, width=0.5, top_left_x=0.05, top_left_y=0.55, theta=23
                ),
                "persimmon_type": "Rojo Brilliante",
                "rjb_quality_options": "Plump",
            },
            {
                "label_ref": "persimmon_014",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.2, width=0.2, top_left_x=0.3, top_left_y=0.2, theta=23
                ),
                "persimmon_type": "Hachiya",
                "hachiya_quality_options": "Plump, Juicy, Large",
            },
            {
                "label_ref": "persimmon_017",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.2, width=0.1, top_left_x=0.6, top_left_y=0.2, theta=23
                ),
                "persimmon_type": "Other persimmon type",
                "Specify persimmon type": "Hiro",
            },
        ],
        105: [
            {
                "label_ref": "persimmon_016",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.5, width=0.5, top_left_x=0.05, top_left_y=0.55, theta=23
                ),
                "persimmon_type": "Rojo Brilliante",
                "rjb_quality_options": "Plump",
            },
            {
                "label_ref": "persimmon_014",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.2, width=0.2, top_left_x=0.3, top_left_y=0.2, theta=23
                ),
                "persimmon_type": "Hachiya",
                "hachiya_quality_options": "Plump, Juicy, Large",
            },
            {
                "label_ref": "persimmon_017",
                "coordinates": RotatableBoundingBoxCoordinates(
                    height=0.2, width=0.1, top_left_x=0.6, top_left_y=0.2, theta=23
                ),
                "persimmon_type": "Other persimmon type",
                "Specify persimmon type": "Maru",
            },
        ],
    },
}

# Cache label rows after initialization
label_row_map = {}

# Step 1: Initialize all label rows with a bundle
with project.create_bundle(bundle_size=BUNDLE_SIZE) as bundle:
    for data_unit in video_frame_labels.keys():
        label_rows = project.list_label_rows_v2(data_title_eq=data_unit)
        assert isinstance(label_rows, list), f"Expected list of label rows for '{data_unit}', got {type(label_rows)}"
        assert label_rows, f"No label row found for {data_unit}"

        label_row = label_rows[0]
        assert label_row is not None, f"Label row is None for {data_unit}"

        label_row.initialise_labels(bundle=bundle)
        assert label_row.ontology_structure is not None, f"Ontology not initialized for label row: {data_unit}"

        label_row_map[data_unit] = label_row  # Cache it for processing later

# Step 2: Process labels and collect label rows to save
label_rows_to_save = []

for data_unit, frame_coordinates in video_frame_labels.items():
    label_row = label_row_map.get(data_unit)
    assert label_row is not None, f"Missing initialized label row for {data_unit}"

    object_instances_by_label_ref = {}

    for frame_number, items in frame_coordinates.items():
        assert isinstance(frame_number, int), f"Frame number must be int, got {type(frame_number)}"

        if not isinstance(items, list):
            items = [items]

        for item in items:
            label_ref = item["label_ref"]
            coord = item["coordinates"]
            persimmon_type = item["persimmon_type"]

            assert persimmon_type in {
                "Hachiya",
                "Fuyu",
                "Rojo Brilliante",
                "Other persimmon type",
            }, f"Unexpected persimmon type '{persimmon_type}' in {data_unit}"

            if label_ref not in object_instances_by_label_ref:
                rbbox_object_instance: ObjectInstance = rbbox_ontology_object.create_instance()
                assert rbbox_object_instance is not None, f"Failed to create ObjectInstance for {label_ref}"

                checklist_attribute = None
                quality_options = []

                # Set persimmon type and checklist attributes
                if persimmon_type == "Hachiya":
                    assert hachiya_option is not None, "Missing 'hachiya_option'"
                    rbbox_object_instance.set_answer(attribute=persimmon_type_radio_attribute, answer=hachiya_option)
                    checklist_attribute = hachiya_checklist_attribute
                    quality_options = item.get("hachiya_quality_options", "").split(", ")
                elif persimmon_type == "Fuyu":
                    assert fuyu_option is not None, "Missing 'fuyu_option'"
                    rbbox_object_instance.set_answer(attribute=persimmon_type_radio_attribute, answer=fuyu_option)
                    checklist_attribute = fuyu_checklist_attribute
                    quality_options = item.get("fuyu_quality_options", "").split(", ")
                elif persimmon_type == "Rojo Brilliante":
                    assert rjb_option is not None, "Missing 'rjb_option'"
                    rbbox_object_instance.set_answer(attribute=persimmon_type_radio_attribute, answer=rjb_option)
                    checklist_attribute = rjb_checklist_attribute
                    quality_options = item.get("rjb_quality_options", "").split(", ")
                elif persimmon_type == "Other persimmon type":
                    assert other_persimmon_option is not None, "Missing 'other_persimmon_option'"
                    rbbox_object_instance.set_answer(
                        attribute=persimmon_type_radio_attribute, answer=other_persimmon_option
                    )
                    text_answer = item.get("Specify persimmon type", "")
                    assert isinstance(text_answer, str), "'Specify persimmon type' must be a string"
                    rbbox_object_instance.set_answer(
                        attribute=other_persimmon_option_text_attribute, answer=text_answer
                    )
                    quality_options = []

                # Set checklist answers based on quality options
                checklist_answers = []
                for quality in quality_options:
                    option = None
                    if quality == "Plump":
                        option = (
                            hachiya_plump_option
                            if persimmon_type == "Hachiya"
                            else fuyu_plump_option
                            if persimmon_type == "Fuyu"
                            else rjb_plump_option
                            if persimmon_type == "Rojo Brilliante"
                            else None
                        )
                    elif quality == "Juicy":
                        option = (
                            hachiya_juicy_option
                            if persimmon_type == "Hachiya"
                            else fuyu_juicy_option
                            if persimmon_type == "Fuyu"
                            else rjb_juicy_option
                            if persimmon_type == "Rojo Brilliante"
                            else None
                        )
                    elif quality == "Large":
                        option = (
                            hachiya_large_option
                            if persimmon_type == "Hachiya"
                            else fuyu_large_option
                            if persimmon_type == "Fuyu"
                            else rjb_large_option
                            if persimmon_type == "Rojo Brilliante"
                            else None
                        )

                    if option:
                        checklist_answers.append(option)
                    else:
                        assert persimmon_type == "Other persimmon type", (
                            f"Invalid quality '{quality}' for persimmon type '{persimmon_type}'"
                        )

                if checklist_attribute and checklist_answers:
                    rbbox_object_instance.set_answer(
                        attribute=checklist_attribute, answer=checklist_answers, overwrite=True
                    )

                object_instances_by_label_ref[label_ref] = rbbox_object_instance

            else:
                rbbox_object_instance = object_instances_by_label_ref[label_ref]

            # Assign coordinates for this frame
            rbbox_object_instance.set_for_frames(coordinates=coord, frames=frame_number)

    # Add object instances to label_row **only if they have frames assigned**
    for rbbox_object_instance in object_instances_by_label_ref.values():
        assert isinstance(rbbox_object_instance, ObjectInstance), "Expected ObjectInstance type"
        if rbbox_object_instance.get_annotation_frames():
            label_row.add_object_instance(rbbox_object_instance)

    # Collect for saving in the next bundle
    label_rows_to_save.append(label_row)

# Step 3: Save all label rows with a bundle
with project.create_bundle(bundle_size=BUNDLE_SIZE) as bundle:
    for label_row in label_rows_to_save:
        assert label_row is not None, "Trying to save a None label row"
        label_row.save(bundle=bundle)
        print(f"Saved label row for {label_row.data_title}")

print("Labels with persimmon type radio buttons, checklist attributes, and text labels added for all data units.")



"""
Code Block Name: Copy Project - Advanced
"""

# Import dependencies
from encord import EncordUserClient
from encord.orm.project import CopyDatasetAction, CopyDatasetOptions, CopyLabelsOptions, ReviewApprovalState

# User input
SSH_PATH = "/Users/chris-encord/ssh-private-key.txt"
PROJECT_ID = "00000000-0000-0000-0000-000000000000"
NEW_PROJECT_TITLE = "My New Project 001"  # Specify the new Project's title
NEW_PROJECT_DECSCRIPTION = "This new Project is for use with ACME Co."  # Specify the new Project's description
NEW_DATASET_TITLE = "My New Dataset 001"  # Specify the new Dataset's title
NEW_DATASET_DESCRIPTION = "This new Dataset is for use with My New Project 001"  # Specify the new Dataset's description
DATASET_ID_01 = "00000000-0000-0000-0000-000000000000"  # Specify a Dataset belonging to the Project being copied
DATASET_ID_02 = "00000000-0000-0000-0000-000000000000"  # Specify a Dataset belonging to the Project being copied
DATA_UNIT_ID_01 = (
    "00000000-0000-0000-0000-000000000000"  # Specify a data_hash for a data unit belonging to a Dataset being copied
)
DATA_UNIT_ID_02 = (
    "00000000-0000-0000-0000-000000000000"  # Specify a data_hash for a data unit belonging to a Dataset being copied
)
DATA_UNIT_ID_03 = (
    "00000000-0000-0000-0000-000000000000"  # Specify a data_hash for a data unit belonging to a Dataset being copied
)
DATA_UNIT_ID_04 = (
    "00000000-0000-0000-0000-000000000000"  # Specify a data_hash for a data unit belonging to a Dataset being copied
)
DATA_UNIT_ID_05 = (
    "00000000-0000-0000-0000-000000000000"  # Specify a data_hash for a data unit belonging to a Dataset being copied
)
DATA_UNIT_ID_06 = (
    "00000000-0000-0000-0000-000000000000"  # Specify a data_hash for a data unit belonging to a Dataset being copied
)
LABEL_ID_01 = "DX2mzwdT"  # Specify an objectHash for a label/classification on a data unit being copied
LABEL_ID_02 = "WNkfZ/7u"  # Specify an objectHash for a label/classification on a data unit being copied


# Create user client using SSH key
user_client: EncordUserClient = EncordUserClient.create_with_ssh_private_key(
    ssh_private_key_path=SSH_PATH,
    # For US platform users use "https://api.us.encord.com"
    domain="https://api.encord.com",
)

# Open the project you want to work on by specifying the Project ID
project = user_client.get_project(PROJECT_ID)

# Copy the project
project.copy_project(
    new_title=NEW_PROJECT_TITLE,
    new_description=NEW_PROJECT_DECSCRIPTION,
    copy_collaborators=True,  # Specify whether Project collaborators are copied to the new Project
    copy_datasets=CopyDatasetOptions(
        action=CopyDatasetAction.CLONE,  # This also creates a new Dataset
        dataset_title=NEW_DATASET_TITLE,
        dataset_description=NEW_DATASET_DESCRIPTION,
        datasets_to_data_hashes_map={
            DATASET_ID_01: [DATA_UNIT_ID_01, DATA_UNIT_ID_02, DATA_UNIT_ID_03],
            DATASET_ID_02: [DATA_UNIT_ID_04, DATA_UNIT_ID_05, DATA_UNIT_ID_06],
        },
    ),
    copy_labels=CopyLabelsOptions(
        accepted_label_statuses=[ReviewApprovalState.APPROVED],  # Copy all labels in the 'Approved' state.
        accepted_label_hashes=[LABEL_ID_01, LABEL_ID_02],  # Copy labels with the listed IDs.
    ),
)



import numpy as np
import pytest

from encord.common.bitmask_operations import serialise_bitmask
from encord.objects.coordinates import PointCoordinate, PolygonCoordinates
from encord.utilities.coco.polygon_utils import find_contours, rle_to_polygons_coordinates


@pytest.fixture
def polygon_with_hole() -> np.ndarray:
    array = np.zeros((100, 100), dtype=np.uint8)
    # Draw outer square (50x50 white square)
    array[25:75, 25:75] = 255
    # Draw inner square (hole - 20x20 black square inside the white one)
    array[40:60, 40:60] = 0

    return array


@pytest.fixture
def two_disjoint_polygons() -> np.ndarray:
    array = np.zeros((100, 100), dtype=np.uint8)
    # Draw two 20x20 white squares
    array[10:30, 10:30] = 255
    array[40:60, 40:60] = 255

    return array


def test_find_contours_with_holes(polygon_with_hole: np.ndarray) -> None:
    polygons = find_contours(polygon_with_hole)
    assert len(polygons) == 1, "Should have one polygon"
    assert polygons == [
        # the hole is a bit "jagged" from the pixelisation, but it's correct.
        [[25, 25, 25, 74, 74, 74, 74, 25], [39, 40, 40, 39, 59, 39, 60, 40, 60, 59, 59, 60, 40, 60, 39, 59]]
    ]


def test_find_contours_multipolygons(two_disjoint_polygons: np.ndarray) -> None:
    polygons = find_contours(two_disjoint_polygons)
    assert len(polygons) == 2, "Should have two polygons"
    assert polygons == [[[40, 40, 40, 59, 59, 59, 59, 40]], [[10, 10, 10, 29, 29, 29, 29, 10]]]


def test_rle_to_polygons_coordinates(polygon_with_hole: np.ndarray) -> None:
    counts = serialise_bitmask(polygon_with_hole.tobytes())
    res = rle_to_polygons_coordinates(counts=counts, height=100, width=100)
    assert isinstance(res, PolygonCoordinates)
    assert res.polygons == [
        [
            [
                PointCoordinate(x=0.25, y=0.25),
                PointCoordinate(x=0.25, y=0.74),
                PointCoordinate(x=0.74, y=0.74),
                PointCoordinate(x=0.74, y=0.25),
            ],
            [
                PointCoordinate(x=0.39, y=0.4),
                PointCoordinate(x=0.4, y=0.39),
                PointCoordinate(x=0.59, y=0.39),
                PointCoordinate(x=0.6, y=0.4),
                PointCoordinate(x=0.6, y=0.59),
                PointCoordinate(x=0.59, y=0.6),
                PointCoordinate(x=0.4, y=0.6),
                PointCoordinate(x=0.39, y=0.59),
            ],
        ]
    ]



"""---
title: "Objects - Ontology Labels"
slug: "sdk-ref-objects-ont-labels"
hidden: false
metadata:
  title: "Objects - Ontology Labels"
  description: "Encord SDK Objects - Ontology Labels."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

import logging
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Type, Union
from uuid import UUID

from encord.client import EncordClientProject
from encord.client import LabelRow as OrmLabelRow
from encord.common.range_manager import RangeManager
from encord.common.time_parser import format_datetime_to_long_string_optional, parse_datetime_optional
from encord.constants.enums import DataType, is_geometric
from encord.exceptions import LabelRowError, WrongProjectTypeError
from encord.http.bundle import Bundle, BundleResultHandler, BundleResultMapper, bundled_operation
from encord.http.limits import (
    LABEL_ROW_BUNDLE_CREATE_LIMIT,
    LABEL_ROW_BUNDLE_GET_LIMIT,
)
from encord.objects import Shape
from encord.objects.attributes import Attribute
from encord.objects.bundled_operations import (
    BundledCreateRowsPayload,
    BundledGetRowsPayload,
    BundledGetStorageItemPayload,
    BundledSaveRowsPayload,
    BundledSetPriorityPayload,
    BundledWorkflowCompletePayload,
    BundledWorkflowReopenPayload,
)
from encord.objects.classification import Classification
from encord.objects.classification_instance import ClassificationInstance, _verify_non_geometric_classifications_range
from encord.objects.constants import (  # pylint: disable=unused-import # for backward compatibility
    DATETIME_LONG_STRING_FORMAT,
    DEFAULT_CONFIDENCE,
    DEFAULT_MANUAL_ANNOTATION,
)
from encord.objects.coordinates import (
    AudioCoordinates,
    BitmaskCoordinates,
    BoundingBoxCoordinates,
    Coordinates,
    CuboidCoordinates,
    HtmlCoordinates,
    PointCoordinate,
    PolygonCoordinates,
    PolygonCoordsToDict,
    PolylineCoordinates,
    RotatableBoundingBoxCoordinates,
    SkeletonCoordinates,
    TextCoordinates,
    Visibility,
)
from encord.objects.frames import Frames, Range, Ranges, frames_class_to_frames_list, frames_to_ranges
from encord.objects.html_node import HtmlRange
from encord.objects.metadata import DataGroupMetadata, DICOMSeriesMetadata, DICOMSliceMetadata
from encord.objects.ontology_object import Object
from encord.objects.ontology_object_instance import ObjectInstance
from encord.objects.ontology_structure import OntologyStructure
from encord.objects.utils import _lower_snake_case
from encord.ontology import Ontology
from encord.orm import storage as orm_storage
from encord.orm.label_row import (
    AnnotationTaskStatus,
    LabelRowMetadata,
    LabelStatus,
    WorkflowGraphNode,
)
from encord.storage import STORAGE_BUNDLE_CREATE_LIMIT, StorageItem, StorageItemInaccessible
from encord.utilities.type_utilities import exhaustive_guard

log = logging.getLogger(__name__)

OntologyTypes = Union[Type[Object], Type[Classification]]
OntologyClasses = Union[Object, Classification]


class LabelRowV2:
    """This class represents a single label row. It corresponds to exactly one data row within a project and holds all
    the labels for that data row.

    You can access many metadata fields with this class directly. To read or write labels, you need to call
    :meth:`.initialise_labels()` first. To upload your added labels, call :meth:`.save()`.
    """

    def __init__(
        self,
        label_row_metadata: LabelRowMetadata,
        project_client: EncordClientProject,
        ontology: Ontology,
    ) -> None:
        self._project_client = project_client
        self._ontology = ontology

        self._label_row_read_only_data: LabelRowV2.LabelRowReadOnlyData = self._parse_label_row_metadata(
            label_row_metadata
        )

        self._is_labelling_initialised = False

        self._frame_to_hashes: defaultdict[int, Set[str]] = defaultdict(set)
        # ^ frames to object and classification hashes

        self._metadata: Optional[Union[DICOMSeriesMetadata, DataGroupMetadata]] = None
        self._frame_metadata: defaultdict[int, Optional[DICOMSliceMetadata]] = defaultdict(lambda: None)

        self._classifications_to_frames: defaultdict[Classification, Set[int]] = defaultdict(set)
        self._classifications_to_ranges: defaultdict[Classification, RangeManager] = defaultdict(RangeManager)

        self._objects_map: Dict[str, ObjectInstance] = dict()
        self._classifications_map: Dict[str, ClassificationInstance] = dict()
        # ^ conveniently a dict is ordered in Python. Use this to our advantage to keep the labels in order
        # at least at the final objects_index/classifications_index level.

        self._storage_item: Optional[StorageItem] = None

    @property
    def label_hash(self) -> Optional[str]:
        """Returns the hash of the label row.

        Returns:
            Optional[str]: The label row hash or None if not set.
        """
        return self._label_row_read_only_data.label_hash

    @property
    def branch_name(self) -> str:
        """Returns the branch name of the label row.

        Returns:
            str: The branch name.
        """
        return self._label_row_read_only_data.branch_name

    @property
    def data_hash(self) -> str:
        """Returns the hash of the data row.

        Returns:
            str: The data row hash.
        """
        return self._label_row_read_only_data.data_hash

    @property
    def group_hash(self) -> Optional[str]:
        """Returns the group hash of the data row.

        Only present if this label row is a child of a data group

        Returns:
            Optional[str]: The data group hash.
        """
        return self._label_row_read_only_data.group_hash

    @property
    def dataset_hash(self) -> str:
        """Returns the hash of the dataset.

        Returns:
            str: The dataset hash.
        """
        return self._label_row_read_only_data.dataset_hash

    @property
    def dataset_title(self) -> str:
        """Returns the title of the dataset.

        Returns:
            str: The dataset title.
        """
        return self._label_row_read_only_data.dataset_title

    @property
    def data_title(self) -> str:
        """Returns the title of the data row.

        Returns:
            str: The data row title.
        """
        return self._label_row_read_only_data.data_title

    @property
    def data_type(self) -> DataType:
        """Returns the data type of the label row.

        Returns:
            DataType: The data type.
        """
        return self._label_row_read_only_data.data_type

    @property
    def file_type(self) -> str | None:
        """Returns the file type of the data row.

        Returns:
            str | None: The file type or None if not set.
        """
        return self._label_row_read_only_data.file_type

    @property
    def client_metadata(self) -> dict | None:
        """Returns the client metadata associated with the label row.

        Returns:
            dict | None: The client metadata or None if not set.
        """
        return self._label_row_read_only_data.client_metadata

    @property
    def label_status(self) -> LabelStatus:
        """Returns the current labeling status for the label row.

        **Note**: This method is not supported for workflow-based projects. Please see our
        :ref:`workflow documentation <tutorials/workflows:Workflows>` for more details.

        Returns:
            LabelStatus: The current labeling status.

        Raises:
            WrongProjectTypeError: If used with workflow-based projects.
        """
        if self.__is_tms2_project:
            raise WrongProjectTypeError(
                '"label_status" property returns incorrect results for workflow-based projects.\
             Please use "workflow_graph_node" property instead.'
            )

        return self._label_row_read_only_data.label_status

    @property
    def annotation_task_status(self) -> AnnotationTaskStatus:
        """Returns the current annotation task status for the label row.

        **Note**: This method is not supported for workflow-based projects. Please see our
        :ref:`workflow documentation <tutorials/workflows:Workflows>` for more details.

        Returns:
            AnnotationTaskStatus: The current annotation task status.

        Raises:
            WrongProjectTypeError: If used with workflow-based projects.
        """
        if self.__is_tms2_project:
            raise WrongProjectTypeError(
                '"annotation_task_status" property returns incorrect results for workflow-based projects.\
             Please use "workflow_graph_node" property instead.'
            )

        assert self._label_row_read_only_data.annotation_task_status is not None  # Never None for Workflow projects

        return self._label_row_read_only_data.annotation_task_status

    @property
    def workflow_graph_node(self) -> Optional[WorkflowGraphNode]:
        """Returns the workflow graph node associated with the label row.

        Returns:
            Optional[WorkflowGraphNode]: The workflow graph node or None if not set.
        """
        return self._label_row_read_only_data.workflow_graph_node

    @property
    def is_shadow_data(self) -> bool:
        """Indicates if the label row is shadow data.

        Returns:
            bool: True if the label row is shadow data, False otherwise.
        """
        return self._label_row_read_only_data.is_shadow_data

    @property
    def created_at(self) -> Optional[datetime]:
        """Returns the creation date of the label row.

        Returns:
            Optional[datetime]: The creation date or None if not yet created.
        """
        return self._label_row_read_only_data.created_at

    @property
    def last_edited_at(self) -> Optional[datetime]:
        """Returns the last edited date of the label row.

        Returns:
            Optional[datetime]: The last edited date or None if not yet edited.
        """
        return self._label_row_read_only_data.last_edited_at

    @property
    def number_of_frames(self) -> int:
        """Returns the number of frames in the label row.

        Returns:
            int: The number of frames.
        """
        return self._label_row_read_only_data.number_of_frames

    @property
    def duration(self) -> Optional[float]:
        """Returns the duration of the video data type.

        Returns:
            Optional[float]: The duration or None if not applicable.
        """
        return self._label_row_read_only_data.duration

    @property
    def fps(self) -> Optional[float]:
        """Returns the frames per second (fps) of the video data type.

        Returns:
            Optional[float]: The fps or None if not applicable.
        """
        return self._label_row_read_only_data.fps

    @property
    def data_link(self) -> Optional[str]:
        """Returns the data link to the underlying object in either cloud storage or Encord storage. This will be `None`
        for DICOM series or image groups created without performance optimisations, as there is no single underlying
        file for these data types.

        This property will contain a signed URL if :meth:`.initialise_labels` was called with `include_signed_url=True`.

        Returns:
            Optional[str]: The data link or None if not applicable.
        """
        return self._label_row_read_only_data.data_link

    @property
    def width(self) -> Optional[int]:
        """Returns the width of the image data type.

        This is `None` for image groups without performance optimisation, as there is no single underlying width
        for this data type.

        Returns:
            Optional[int]: The width or None if not applicable.
        """
        return self._label_row_read_only_data.width

    @property
    def height(self) -> Optional[int]:
        """Returns the height of the image data type.

        This is `None` for image groups without performance optimisation, as there is no single underlying height
        for this data type.

        Returns:
            Optional[int]: The height or None if not applicable.
        """
        return self._label_row_read_only_data.height

    @property
    def audio_codec(self) -> Optional[str]:
        """Returns the codec of the audio data type.

        This only applies to audio data types, and returns None for all other data types.

        Returns:
            Optional[str]: The codec or None if not applicable.
        """
        return self._label_row_read_only_data.audio_codec

    @property
    def audio_sample_rate(self) -> Optional[int]:
        """Returns the sample rate of the audio data type.

        This only applies to audio data types, and returns None for all other data types.

        Returns:
            Optional[int]: The sample rate or None if not applicable.
        """
        return self._label_row_read_only_data.audio_sample_rate

    @property
    def audio_bit_depth(self) -> Optional[int]:
        """Returns the bit depth of the audio data type.

        This only applies to audio data types, and returns None for all other data types.

        Returns:
            Optional[int]: The bit depth or None if not applicable.
        """
        return self._label_row_read_only_data.audio_bit_depth

    @property
    def audio_num_channels(self) -> Optional[int]:
        """Returns the number of channels of the audio data type.

        This only applies to audio data types, and returns None for all other data types.

        Returns:
            Optional[int]: The number of channels or None if not applicable.
        """
        return self._label_row_read_only_data.audio_num_channels

    @property
    def backing_item_uuid(self) -> Optional[UUID]:
        """Returns the unique identifier (UUID) for the backing storage item associated with this label row.

        While it is always included in server responses, it is marked as optional for backward compatibility with earlier versions.

        Returns:
            Optional[UUID]: The backing storage item id or None if not found.
        """
        # TODO: Mark required in 0.2 release
        return self._label_row_read_only_data.backing_item_uuid

    @property
    def task_uuid(self) -> Optional[UUID]:
        """Returns the workflow task uuids for the task associated with the data unit.

        This property only works for workflow-based projects.

        Returns:
            Optional[UUID]: The workflow task uuid or None if not applicable.

        Raises:
            WrongProjectTypeError: If used with non-workflow-based projects.
        """
        if not self.__is_tms2_project:
            raise WrongProjectTypeError('"task_uuid" property only works with workflow-based projects.')

        return self._label_row_read_only_data.task_uuid

    @property
    def priority(self) -> Optional[float]:
        """Returns the workflow priority for the task associated with the data unit.

        This property only works for workflow-based projects.

        It is None for label rows in the "complete" state.

        Returns:
            Optional[float]: The workflow priority or None if not applicable.

        Raises:
            WrongProjectTypeError: If used with non-workflow-based projects.
        """
        if not self.__is_tms2_project:
            raise WrongProjectTypeError('"priority" property only works with workflow-based projects.')

        return self._label_row_read_only_data.priority

    @property
    def is_valid(self) -> bool:
        """Indicates if the labels are valid.

        For labels uploaded via the SDK, a check is run to ensure that the labels are valid.
        This property returns `True` if the labels have correct structure and match the project ontology.

        If it is `False`, then loading the labels via `LabelRowV2` will raise an error, and the label editor
        will not be able to load the labels.

        You can call :meth:`.get_validation_errors` to get the validation error messages.

        Returns:
            bool: True if the labels are valid, False otherwise.
        """
        return self._label_row_read_only_data.is_valid

    @property
    def ontology_structure(self) -> OntologyStructure:
        """Get the corresponding ontology structure."""
        return self._ontology.structure

    @property
    def is_labelling_initialised(self) -> bool:
        """Check if labelling is initialized.

        Returns:
            bool: `True` if labelling is initialized, `False` otherwise.

        If this is `False`, call the :meth:`.initialise_labels` method to read or write specific
        ObjectInstances or ClassificationInstances.
        """
        return self._is_labelling_initialised

    @property
    def __is_tms2_project(self) -> bool:
        return self.workflow_graph_node is not None

    @property
    def assigned_user_email(self) -> str | None:
        """Email of the user assigned to annotation or review task for a workflow-based project.
        In case of completed task, it is None.

        Returns:
            str | None: The email of the user assigned to the data.

        """
        return self._label_row_read_only_data.assigned_user_email

    @property
    def last_actioned_by_user_email(self) -> str | None:
        """Email of the user who last actioned the data.

        Returns:
            str | None: The email of the user who last actioned the data.
        """
        return self._label_row_read_only_data.last_actioned_by_user_email

    def get_storage_item(self, get_signed_urls: bool = False) -> StorageItem:
        """Returns the storage item associated with the label row.
        This function can be used to get storage item details like storage folder, signed url, created at, item type, client metadata, etc.
        """
        if self._label_row_read_only_data.backing_item_uuid is None:
            raise LabelRowError("Storage item is not found for the label row")

        if not self._storage_item:
            self._storage_item = StorageItem._get_item(
                self._project_client._api_client,
                self._label_row_read_only_data.backing_item_uuid,
                get_signed_url=get_signed_urls,
            )

        return self._storage_item

    def initialise_storage_item(self, get_signed_url: bool = False, bundle: Optional[Bundle] = None) -> None:
        """Initialise the storage item associated with the label row.

        This function will download the storage item details from the Encord server.
        if you want to get the signed url, you can set the get_signed_url to True.

        Args:
            get_signed_url: If `True`,is set true, you can get the signed url from LabelRowV2.get_storage_item().get_signed_url()
            bundle: If not provided, initialization is performed independently. If provided,
                initialization is delayed and performed along with other objects in the same bundle.
        """

        if self._label_row_read_only_data.backing_item_uuid is None:
            raise LabelRowError("Storage item is not found for the label row")

        bundled_operation(
            bundle,
            operation=self._project_client._api_client.get_bound_operation(StorageItem._get_items_bulk),
            payload=BundledGetStorageItemPayload(
                item_uuids=[str(self._label_row_read_only_data.backing_item_uuid)],
                get_signed_url=get_signed_url,
            ),
            result_mapper=BundleResultMapper[orm_storage.StorageItem](
                result_mapping_predicate=lambda r: str(r.uuid),
                result_handler=BundleResultHandler(
                    predicate=str(self._label_row_read_only_data.backing_item_uuid), handler=self._set_orm_item
                ),
            ),
            limit=STORAGE_BUNDLE_CREATE_LIMIT,
        )

    def _set_orm_item(self, orm_storage_item: orm_storage.StorageItem) -> None:
        self._storage_item = StorageItem(self._project_client._api_client, orm_storage_item)

    def initialise_labels(
        self,
        include_object_feature_hashes: Optional[Set[str]] = None,
        include_classification_feature_hashes: Optional[Set[str]] = None,
        include_reviews: bool = False,
        overwrite: bool = False,
        bundle: Optional[Bundle] = None,
        *,
        include_signed_url: bool = False,
    ) -> None:
        """Initialize labels from the Encord server.

        This function downloads or exports labels stored on the Encord server and performs
        any other reading or writing operations. If you only want to inspect a subset of labels,
        you can filter them. Note that filtering and later uploading will delete all previously
        filtered labels.

        If the label is not in progress, this sets the label status to `LabelStatus.LABEL_IN_PROGRESS`.

        Args:
            include_object_feature_hashes: If `None`, all objects will be included. Otherwise,
                only objects with the specified feature hashes will be included.
                WARNING: Use this filter only for reading labels. Saving a filtered subset will
                delete all other object instances stored in the Encord platform.
            include_classification_feature_hashes: If `None`, all classifications will be included.
                Otherwise, only classifications with the specified feature hashes will be included.
                WARNING: Use this filter only for reading labels. Saving a filtered subset will
                delete all other classification instances stored in the Encord platform.
            include_reviews: Whether to request read-only information about the reviews of the label row.
            overwrite: If `True`, overwrite current labels with those stored in the Encord server.
                If `False` and the label row is already initialized, this function will raise an error.
            bundle: If not provided, initialization is performed independently. If provided,
                initialization is delayed and performed along with other objects in the same bundle.
            include_signed_url: If `True`, the :attr:`.data_link` property will contain a signed URL.
                See documentation for :attr:`.data_link` for more details.
        """
        if self.is_labelling_initialised and not overwrite:
            raise LabelRowError(
                "You are trying to re-initialise a label row that has already been initialised. This would overwrite "
                "current labels. If this is your intend, set the `overwrite` flag to `True`."
            )

        if not self.label_hash:
            # If label_hash is None, it means we need to explicitly create the label row first
            bundled_operation(
                bundle,
                operation=self._project_client.create_label_rows,
                payload=BundledCreateRowsPayload(
                    uids=[self.data_hash], get_signed_url=include_signed_url, branch_name=self.branch_name
                ),
                result_mapper=BundleResultMapper[OrmLabelRow](
                    result_mapping_predicate=lambda r: r["data_hash"],
                    result_handler=BundleResultHandler(predicate=self.data_hash, handler=self.from_labels_dict),
                ),
                limit=LABEL_ROW_BUNDLE_CREATE_LIMIT,
            )
        else:
            bundled_operation(
                bundle,
                operation=self._project_client.get_label_rows,
                payload=BundledGetRowsPayload(
                    uids=[self.label_hash],
                    get_signed_url=include_signed_url,
                    include_object_feature_hashes=include_object_feature_hashes,
                    include_classification_feature_hashes=include_classification_feature_hashes,
                    include_reviews=include_reviews,
                ),
                result_mapper=BundleResultMapper[OrmLabelRow](
                    result_mapping_predicate=lambda r: r["label_hash"],
                    result_handler=BundleResultHandler(predicate=self.label_hash, handler=self.from_labels_dict),
                ),
                limit=LABEL_ROW_BUNDLE_GET_LIMIT,
            )

    def from_labels_dict(self, label_row_dict: dict) -> None:
        """Initialize the LabelRow from a label row dictionary.

        This function also initializes the label row. It resets all the labels currently
        stored within this class.

        Args:
            label_row_dict: The dictionary of all labels in the Encord format.
        """
        self._is_labelling_initialised = True

        self._label_row_read_only_data = self._parse_label_row_dict(label_row_dict)
        self._frame_to_hashes = defaultdict(set)
        self._classifications_to_frames = defaultdict(set)
        self._classifications_to_ranges = defaultdict(RangeManager)

        self._metadata = None
        self._frame_metadata = defaultdict(lambda: None)

        self._objects_map = dict()
        self._classifications_map = dict()
        self._parse_labels_from_dict(label_row_dict)

    def get_image_hash(self, frame_number: int) -> Optional[str]:
        """Get the corresponding image hash for the frame number.

        Args:
            frame_number: The frame number.

        Returns:
            Optional[str]: The image hash if the frame number is within bounds, `None` otherwise.

        Raises:
            LabelRowError: If this function is used for non-image data types.
        """
        self._check_labelling_is_initalised()

        if self.data_type not in (DataType.IMAGE, DataType.IMG_GROUP):
            raise LabelRowError("This function is only supported for label rows of image or image group data types.")

        return self._label_row_read_only_data.frame_to_image_hash.get(frame_number)

    def get_frame_number(self, image_hash: str) -> Optional[int]:
        """Get the corresponding frame number for the image hash.

        Args:
            image_hash: The image hash.

        Returns:
            Optional[int]: The frame number if the image hash is found, `None` otherwise.

        Raises:
            LabelRowError: If this function is used for non-image data types.
        """
        self._check_labelling_is_initalised()

        if self.data_type not in (DataType.IMAGE, DataType.IMG_GROUP):
            raise LabelRowError("This function is only supported for label rows of image or image group data types.")
        return self._label_row_read_only_data.image_hash_to_frame[image_hash]

    def save(self, bundle: Optional[Bundle] = None, validate_before_saving: bool = False) -> None:
        """Upload the created labels to the Encord server.

        This will overwrite any labels that someone else has created on the platform in the meantime.

        Args:
            bundle: If not provided, save is executed immediately. If provided, save is executed
                as part of the bundle.
            validate_before_saving: Enable stricter server-side integrity checks. Default is `False`.
        """
        self._check_labelling_is_initalised()
        assert self.label_hash is not None  # Checked earlier, assert is just to silence mypy

        bundled_operation(
            bundle,
            self._project_client.save_label_rows,
            payload=BundledSaveRowsPayload(
                uids=[self.label_hash], payload=[self.to_encord_dict()], validate_before_saving=validate_before_saving
            ),
        )

    @property
    def metadata(self) -> Optional[Union[DICOMSeriesMetadata, DataGroupMetadata]]:
        """Get metadata for the given data type.

        Returns:
            Optional[DICOMSeriesMetadata]: Metadata for the data type, or `None` if not supported.

        Currently only supported for DICOM. Returns `None` for other formats.

        Label row needs to be initialized before using this property.
        """
        self._check_labelling_is_initalised()
        return self._metadata

    def get_frame_view(self, frame: Union[int, str] = 0) -> FrameView:
        """Get a view of the specified frame.

        Args:
            frame: Either the frame number or the image hash if the data type is an image or image group.
                Defaults to the first frame.

        Returns:
            FrameView: A view of the specified frame.
        """
        self._method_not_supported_for_audio()

        self._check_labelling_is_initalised()
        if isinstance(frame, str):
            frame_num = self.get_frame_number(frame)
            if frame_num is None:
                raise LabelRowError(f"Image hash {frame} not found in the label row")
        else:
            frame_num = frame

        return self.FrameView(self, self._label_row_read_only_data, frame_num)

    def _get_frame_metadata_list(self) -> List[LabelRowReadOnlyDataImagesDataEntry]:
        if self._label_row_read_only_data.data_type in [DataType.IMAGE, DataType.VIDEO]:
            return [
                self.LabelRowReadOnlyDataImagesDataEntry(
                    index=0,
                    title=self._label_row_read_only_data.data_title,
                    file_type=self._label_row_read_only_data.file_type or "",
                    height=self._label_row_read_only_data.height or 0,
                    width=self._label_row_read_only_data.width or 0,
                    image_hash=self._label_row_read_only_data.data_hash,
                )
            ]

        images_data = self._label_row_read_only_data.images_data
        if images_data is None:
            raise LabelRowError("Image data is not present in the label row")
        return images_data

    def get_frame_metadata(self, frame: Union[int, str] = 0) -> FrameViewMetadata:
        """Get metadata for a specific frame or image hash.

        Args:
            frame: The frame number or the image hash. Defaults to the first frame.

        Returns:
            FrameViewMetadata: Metadata for the specified frame or image hash.

        Raises:
            LabelRowError: If the specified frame or image hash is not found in the label row.
        """
        self._method_not_supported_for_audio()

        images_data = self._get_frame_metadata_list()
        if isinstance(frame, str):
            data_meta = None
            for data in images_data:
                if data.image_hash == frame:
                    data_meta = data
                    break
            if data_meta is None:
                raise LabelRowError(f"Image hash {frame} not found in the label row")
        else:
            data_meta = None
            for data in images_data:
                if data.index == frame:
                    data_meta = data
                    break
            if data_meta is None:
                raise LabelRowError(f"Frame {frame} not found in the label row")

        return self.FrameViewMetadata(data_meta)

    def get_frames_metadata(self) -> List[FrameViewMetadata]:
        """Get metadata for all frames in the image group.

        Returns:
            List[FrameViewMetadata]: A list of metadata for each frame.
        """
        views = []

        images_data = self._get_frame_metadata_list()

        for data in images_data:
            views.append(self.FrameViewMetadata(data))
        return views

    def get_frame_views(self) -> List[FrameView]:
        """Get views for all frames.

        Returns:
            List[FrameView]: A list of frame views in order of available frames.
        """
        self._check_labelling_is_initalised()
        ret = []
        for frame in range(self.number_of_frames):
            ret.append(self.get_frame_view(frame))
        return ret

    def get_object_instances(
        self, filter_ontology_object: Optional[Object] = None, filter_frames: Optional[Frames] = None
    ) -> List[ObjectInstance]:
        """Get all object instances that match the given filters.

        Args:
            filter_ontology_object: Optionally filter by a specific ontology object.
            filter_frames: Optionally filter by specific frames.

        Returns:
            List[ObjectInstance]: A list of `ObjectInstance`s that match the filters.
        """
        self._check_labelling_is_initalised()

        ret: List[ObjectInstance] = list()

        if filter_frames is not None:
            filtered_frames_list = frames_class_to_frames_list(filter_frames)
        else:
            filtered_frames_list = list()

        for object_ in self._objects_map.values():
            # filter by ontology object
            if not (
                filter_ontology_object is None
                or object_.ontology_item.feature_node_hash == filter_ontology_object.feature_node_hash
            ):
                continue

            # filter by frame
            if filter_frames is None:
                append = True
            else:
                append = False
            for frame in filtered_frames_list:
                hashes = self._frame_to_hashes.get(frame, set())
                if object_.object_hash in hashes:
                    append = True
                    break

            if append:
                ret.append(object_)

        return ret

    def add_object_instance(self, object_instance: ObjectInstance, force: bool = True) -> None:
        """Add an object instance to the label row. If the object instance already exists, it overwrites the current instance.

        Args:
            object_instance: The object instance to add.
            force: If `True`, overwrites the current objects; otherwise, it will replace the current object.

        Raises:
            LabelRowError: If the object instance is already part of another LabelRowV2.
        """
        self._method_not_supported_for_audio(range_only=object_instance.is_range_only())

        self._check_labelling_is_initalised()

        object_instance.is_valid()

        # We want to ensure that we are only adding the object_instance to a label_row
        # IF AND ONLY IF the file type is text/html and the object_instance has range_html set
        if self.file_type == "text/html" and object_instance.range_html is None:
            raise LabelRowError(
                "Unable to assign object instance without a html range to a html file. "
                f"Please ensure the object instance exists on frame=0, and has coordinates of type {HtmlCoordinates}."
            )
        elif self.file_type != "text/html" and object_instance.range_html is not None:
            raise LabelRowError(
                "Unable to assign object instance with a html range to a non-html file. "
                f"Please ensure the object instance does not have coordinates of type {HtmlCoordinates}."
            )

        if object_instance.is_assigned_to_label_row():
            raise LabelRowError(
                "The supplied ObjectInstance is already part of a LabelRowV2. You can only add a ObjectInstance to one "
                "LabelRowV2. You can do a ObjectInstance.copy() to create an identical ObjectInstance which is not part of "
                "any LabelRowV2."
            )

        object_hash = object_instance.object_hash
        if object_hash in self._objects_map and not force:
            raise LabelRowError(
                "The supplied ObjectInstance was already previously added. (the object_hash is the same)."
            )
        elif object_hash in self._objects_map and force:
            self._objects_map.pop(object_hash)

        self._objects_map[object_hash] = object_instance
        object_instance._parent = self

        frames = set(_frame_views_to_frame_numbers(object_instance.get_annotations()))
        self._add_to_frame_to_hashes_map(object_instance, frames)

    def add_classification_instance(self, classification_instance: ClassificationInstance, force: bool = False) -> None:
        """Add a classification instance to the label row.

        Args:
            classification_instance: The classification instance to add.
            force: If `True`, overwrites the current objects; otherwise, it will replace the current object.

        Raises:
            LabelRowError: If the classification instance is already part of another LabelRowV2 or has overlapping frames.
        """
        self._check_labelling_is_initalised()

        classification_instance.is_valid()

        # TODO: Need to update the docstring for this method, talk to Laverne.
        if not classification_instance.is_range_only() and not is_geometric(self.data_type):
            raise LabelRowError(
                f"To add a ClassificationInstance object to a label row where data_type = {self.data_type},"
                "the ClassificationInstance object needs to be created with the "
                "range_only property set to True."
                "You can do ClassificationInstance(range_only=True) or "
                "Classification.create_instance(range_only=True) to achieve this."
            )

        if classification_instance.is_assigned_to_label_row():
            raise LabelRowError(
                "Provided ClassificationInstance object is already attached to a different LabelRowV2 object. "
                "You can only add a ClassificationInstance to one label row. "
                "You can do a ClassificationInstance.copy() to create an identical ClassificationInstance which is not part of any label row."
            )

        classification_hash = classification_instance.classification_hash
        if classification_hash in self._classifications_map and not force:
            raise LabelRowError(
                f"A ClassificationInstance for classification hash '{classification_hash}' already exists on the label row object. "
                f"Pass 'force=True' to override it."
            )

        """
        Implementation here diverges because non-geometric data will operate on ranges, whereas
        everything else will operate on frames.
        """
        if not is_geometric(self.data_type):
            self._add_classification_instance_for_range(
                classification_instance=classification_instance,
                force=force,
            )
        else:
            frames = set(_frame_views_to_frame_numbers(classification_instance.get_annotations()))
            already_present_frames = self._is_classification_already_present_on_frames(
                classification_instance.ontology_item,
                frames,
            )

            if already_present_frames and not force:
                raise LabelRowError(
                    f"A ClassificationInstance '{classification_hash}' was already added and has overlapping frames. "
                    f"Overlapping frames that were found are `{frames_to_ranges(already_present_frames)}`. "
                    f"Make sure that you only add classifications which are on frames where the same type of "
                    f"classification does not yet exist."
                )

            if classification_hash in self._classifications_map and force:
                self._classifications_map.pop(classification_hash)

            self._classifications_map[classification_hash] = classification_instance
            classification_instance._parent = self

            self._classifications_to_frames[classification_instance.ontology_item].update(frames)
            self._add_to_frame_to_hashes_map(classification_instance, frames)

    # This should only be used for Non-geometric data
    def _add_classification_instance_for_range(
        self,
        classification_instance: ClassificationInstance,
        force: bool,
    ):
        classification_hash = classification_instance.classification_hash
        ranges_to_add = classification_instance.range_list

        if classification_instance.is_range_only():
            _verify_non_geometric_classifications_range(ranges_to_add, self)

        already_present_ranges = self._is_classification_already_present_on_ranges(
            classification_instance.ontology_item, ranges_to_add
        )

        if already_present_ranges and not force:
            raise LabelRowError(
                f"A ClassificationInstance '{classification_hash}' was already added and has overlapping frames. "
                f"Overlapping frames that were found are `{already_present_ranges}`. "
                f"Make sure that you only add classifications which are on frames where the same type of "
                f"classification does not yet exist."
            )

        if classification_hash in self._classifications_map and force:
            self._classifications_map.pop(classification_hash)

        self._classifications_map[classification_hash] = classification_instance
        classification_instance._parent = self

        self._classifications_to_ranges[classification_instance.ontology_item].add_ranges(ranges_to_add)

    def remove_classification(self, classification_instance: ClassificationInstance):
        """Remove a classification instance from a label row.

        Args:
            classification_instance: The classification instance to remove.
        """
        self._check_labelling_is_initalised()

        classification_hash = classification_instance.classification_hash
        self._classifications_map.pop(classification_hash)

        if not is_geometric(self.data_type):
            range_manager = self._classifications_to_ranges[classification_instance.ontology_item]
            ranges_to_remove = classification_instance.range_list
            range_manager.remove_ranges(ranges_to_remove)
        else:
            all_frames = self._classifications_to_frames[classification_instance.ontology_item]
            actual_frames = _frame_views_to_frame_numbers(classification_instance.get_annotations())
            for actual_frame in actual_frames:
                all_frames.remove(actual_frame)

    def add_to_single_frame_to_hashes_map(
        self, label_item: Union[ObjectInstance, ClassificationInstance], frame: int
    ) -> None:
        # This is an internal function, it is not meant to be called by the SDK user.
        self._check_labelling_is_initalised()

        if isinstance(label_item, ObjectInstance):
            self._frame_to_hashes[frame].add(label_item.object_hash)
        elif isinstance(label_item, ClassificationInstance):
            self._frame_to_hashes[frame].add(label_item.classification_hash)
        else:
            raise NotImplementedError(f"Got an unexpected label item class `{type(label_item)}`")

    def get_classification_instances(
        self, filter_ontology_classification: Optional[Classification] = None, filter_frames: Optional[Frames] = None
    ) -> List[ClassificationInstance]:
        """Get all classification instances that match the given filters.

        Args:
            filter_ontology_classification: Optionally filter by a specific ontology classification.
            filter_frames: Optionally filter by specific frames.

        Returns:
            List[ClassificationInstance]: All the `ClassificationInstance`s that match the filter.
        """
        self._check_labelling_is_initalised()

        ret: List[ClassificationInstance] = list()

        if filter_frames is not None:
            filtered_frames_list = frames_class_to_frames_list(filter_frames)
        else:
            filtered_frames_list = list()

        for classification in self._classifications_map.values():
            # filter by ontology object
            if not (
                filter_ontology_classification is None
                or classification.ontology_item.feature_node_hash == filter_ontology_classification.feature_node_hash
            ):
                continue

            # filter by frame
            if filter_frames is None:
                append = True
            else:
                append = False
            for frame in filtered_frames_list:
                hashes = self._frame_to_hashes.get(frame, set())
                if classification.classification_hash in hashes:
                    append = True
                    break

            if append:
                ret.append(classification)
        return ret

    def remove_object(self, object_instance: ObjectInstance):
        """Remove an object instance from a label row.

        Args:
            object_instance: The object instance to remove.
        """
        self._check_labelling_is_initalised()

        self._objects_map.pop(object_instance.object_hash)
        if not object_instance.is_range_only():
            self._remove_from_frame_to_hashes_map(
                _frame_views_to_frame_numbers(object_instance.get_annotations()), object_instance.object_hash
            )
        object_instance._parent = None

    def to_encord_dict(self) -> Dict[str, Any]:
        """Convert the label row to a dictionary in Encord format.

        This is an internal helper function. Likely this should not be used by a user. To upload labels use the
        :meth:`.save()` function.

        Returns:
            Dict[str, Any]: A dictionary representing the label row in Encord format.
        """
        self._check_labelling_is_initalised()

        ret: Dict[str, Any] = {}
        read_only_data = self._label_row_read_only_data

        ret["label_hash"] = read_only_data.label_hash
        ret["branch_name"] = read_only_data.branch_name
        ret["created_at"] = format_datetime_to_long_string_optional(read_only_data.created_at)
        ret["last_edited_at"] = format_datetime_to_long_string_optional(read_only_data.last_edited_at)
        ret["data_hash"] = read_only_data.data_hash
        ret["dataset_hash"] = read_only_data.dataset_hash
        ret["dataset_title"] = read_only_data.dataset_title
        ret["data_title"] = read_only_data.data_title
        ret["data_type"] = read_only_data.data_type.value
        ret["annotation_task_status"] = read_only_data.annotation_task_status
        ret["is_shadow_data"] = read_only_data.is_shadow_data
        ret["object_answers"] = self._to_object_answers()
        ret["classification_answers"] = self._to_classification_answers()
        ret["object_actions"] = self._to_object_actions()
        ret["label_status"] = read_only_data.label_status.value
        ret["data_units"] = self._to_encord_data_units()

        return ret

    def workflow_reopen(self, bundle: Optional[Bundle] = None) -> None:
        """Return a label row to the first annotation stage for re-labeling.

        No data will be lost during this call. This method is only relevant for the projects that use the
        :ref:`Workflow <tutorials/workflows:Workflows>` feature, and will raise an error for pre-workflow projects.

        Args:
            bundle: Optional parameter. If passed, the method will be executed in a deferred way as part of the bundle.
        """
        if self.label_hash is None:
            # Label has not yet moved from the initial state, nothing to do
            return

        bundled_operation(
            bundle,
            self._project_client.workflow_reopen,
            payload=BundledWorkflowReopenPayload(label_hashes=[self.label_hash]),
        )

    def workflow_complete(self, bundle: Optional[Bundle] = None) -> None:
        """Move a label row to the final workflow node, marking it as 'Complete'.

        This method can be called only for labels for which :meth:`.initialise_labels()` was called at least once, and
        consequentially the "label_hash" field is not `None`. Labels need not be initialised every time the
        workflow_complete() method is called.

        This method is only relevant for the projects that use the :ref:`Workflow <tutorials/workflows:Workflows>`
        feature, and will raise an error for projects that don't use Workflows.

        Args:
            bundle: Optional parameter. If passed, the method will be executed in a deferred way as part of the bundle.

        Raises:
            LabelRowError: If the label hash is None.
        """
        if self.label_hash is None:
            raise LabelRowError(
                "For this operation you will need to initialise labelling first. Call the .initialise_labels() "
                "to do so first."
            )

        bundled_operation(
            bundle,
            self._project_client.workflow_complete,
            payload=BundledWorkflowCompletePayload(label_hashes=[self.label_hash]),
        )

    def set_priority(self, priority: float, bundle: Optional[Bundle] = None) -> None:
        """Set priority for a task in a workflow project.

        Args:
            priority: Float value from 0.0 to 1.0, where 1.0 is the highest priority.
            bundle: Optional parameter. If passed, the method will be executed in a deferred way as part of the bundle.

        Raises:
            WrongProjectTypeError: If the project is not a workflow-based project.
        """
        if not self.__is_tms2_project:
            raise WrongProjectTypeError("Setting priority only possible for workflow-based projects")

        bundled_operation(
            bundle,
            self._project_client.workflow_set_priority,
            payload=BundledSetPriorityPayload(priorities=[(self.data_hash, priority)]),
        )

    def get_validation_errors(self) -> List[str] | None:
        """Get validation errors for the label row.

        Returns:
            List[str] | None: A list of error messages if the label row is invalid, otherwise `None`.
        """
        if not self.label_hash or self.is_valid:
            return None

        return self._project_client.get_label_validation_errors(self.label_hash)

    class FrameViewMetadata:
        """Class to inspect metadata for a frame view."""

        def __init__(self, images_data: LabelRowV2.LabelRowReadOnlyDataImagesDataEntry):
            self._image_data = images_data

        @property
        def title(self) -> str:
            return self._image_data.title
            """
            Get the title of the image.

            Returns:
                str: The image title.
            """

        @property
        def file_type(self) -> str:
            return self._image_data.file_type
            """
            Get the file type of the image.

            Returns:
                str: The image file type.
            """

        @property
        def width(self) -> int:
            return self._image_data.width
            """
            Get the width of the image.

            Returns:
                int: The image width.
            """

        @property
        def height(self) -> int:
            return self._image_data.height
            """
            Get the height of the image.

            Returns:
                int: The image height.
            """

        @property
        def image_hash(self) -> str:
            """Get the hash of the image.

            Returns:
                str: The image hash.
            """
            return self._image_data.image_hash

        @property
        def frame_number(self) -> int:
            """Get the frame number.

            Returns:
                int: The frame number.
            """
            return self._image_data.index

    class FrameView:
        """This class can be used to inspect object or classification instances on a given frame,
        or metadata such as image file size for a given frame.
        """

        def __init__(
            self, label_row: LabelRowV2, label_row_read_only_data: LabelRowV2.LabelRowReadOnlyData, frame: int
        ):
            self._label_row = label_row
            self._label_row_read_only_data = label_row_read_only_data
            self._frame = frame

        @property
        def image_hash(self) -> str:
            """Get the image hash for the current frame.

            Returns:
                str: The image hash.

            Raises:
                LabelRowError: If the data type is not IMAGE or IMG_GROUP.
            """
            if self._label_row.data_type not in [DataType.IMAGE, DataType.IMG_GROUP]:
                raise LabelRowError("Image hash can only be retrieved for DataType.IMAGE or DataType.IMG_GROUP")
            return self._frame_level_data().image_hash

        @property
        def image_title(self) -> str:
            """Get the image title for the current frame.

            Returns:
                str: The image title.

            Raises:
                LabelRowError: If the data type is not IMAGE or IMG_GROUP.
            """
            if self._label_row.data_type not in [DataType.IMAGE, DataType.IMG_GROUP]:
                raise LabelRowError("Image title can only be retrieved for DataType.IMAGE or DataType.IMG_GROUP")
            return self._frame_level_data().image_title

        @property
        def file_type(self) -> str:
            """Get the file type for the current frame.

            Returns:
                str: The file type.

            Raises:
                LabelRowError: If the data type is not IMAGE or IMG_GROUP.
            """
            if self._label_row.data_type not in [DataType.IMAGE, DataType.IMG_GROUP]:
                raise LabelRowError("File type can only be retrieved for DataType.IMAGE or DataType.IMG_GROUP")
            return self._frame_level_data().file_type

        @property
        def frame(self) -> int:
            """Get the frame number.

            Returns:
                int: The frame number.
            """
            return self._frame

        @property
        def width(self) -> int:
            """Get the width of the image or image group.

            Returns:
                int: The width of the image or image group.

            Raises:
                LabelRowError: If the width is not set for the data type.
            """
            if self._label_row.data_type in [DataType.IMG_GROUP]:
                return self._frame_level_data().width
            elif self._label_row_read_only_data.width is not None:
                return self._label_row_read_only_data.width
            else:
                raise LabelRowError(f"Width is expected but not set for the data type {self._label_row.data_type}")

        @property
        def height(self) -> int:
            """Get the height of the image or image group.

            Returns:
                int: The height of the image or image group.

            Raises:
                LabelRowError: If the height is not set for the data type.
            """
            if self._label_row.data_type in [DataType.IMG_GROUP]:
                return self._frame_level_data().height
            elif self._label_row_read_only_data.height is not None:
                return self._label_row_read_only_data.height
            else:
                raise LabelRowError(f"Height is expected but not set for the data type {self._label_row.data_type}")

        @property
        def data_link(self) -> Optional[str]:
            """Get the data link for the current frame.

            Returns:
                Optional[str]: The data link, or `None` if not available.

            Raises:
                LabelRowError: If the data type is not IMAGE or IMG_GROUP.
            """
            if self._label_row.data_type not in [DataType.IMAGE, DataType.IMG_GROUP]:
                raise LabelRowError("Data link can only be retrieved for DataType.IMAGE or DataType.IMG_GROUP")
            return self._frame_level_data().data_link

        @property
        def metadata(self) -> Optional[DICOMSliceMetadata]:
            """Get annotation metadata for the current frame.

            Returns:
                Optional[DICOMSliceMetadata]: The metadata for the frame, or `None` if not applicable.

            Note:
                Currently only supported for DICOM, and will return `None` for other formats.
            """
            return self._label_row._frame_metadata[self._frame]

        def add_object_instance(
            self,
            object_instance: ObjectInstance,
            coordinates: Coordinates,
            *,
            overwrite: bool = False,
            created_at: Optional[datetime] = None,
            created_by: Optional[str] = None,
            last_edited_at: Optional[datetime] = None,
            last_edited_by: Optional[str] = None,
            confidence: Optional[float] = None,
            manual_annotation: Optional[bool] = None,
        ) -> None:
            """Add an object instance to the current frame.

            Args:
                object_instance: The object instance to add.
                coordinates: The coordinates for the object instance.
                overwrite: Whether to overwrite existing instance data (default is False).
                created_at: Optional creation timestamp.
                created_by: Optional creator identifier.
                last_edited_at: Optional last edited timestamp.
                last_edited_by: Optional last editor identifier.
                confidence: Optional confidence score.
                manual_annotation: Optional flag indicating manual annotation.

            Raises:
                LabelRowError: If the object instance is already assigned to a different label row.
            """
            label_row = object_instance.is_assigned_to_label_row()
            if label_row and self._label_row != label_row:
                raise LabelRowError(
                    "This object instance is already assigned to a different label row. It cannot be "
                    "added to multiple label rows at once."
                )

            object_instance.set_for_frames(
                coordinates,
                self._frame,
                overwrite=overwrite,
                created_at=created_at,
                created_by=created_by,
                last_edited_at=last_edited_at,
                last_edited_by=last_edited_by,
                confidence=confidence,
                manual_annotation=manual_annotation,
            )

            if not label_row:
                self._label_row.add_object_instance(object_instance)

        def add_classification_instance(
            self,
            classification_instance: ClassificationInstance,
            *,
            overwrite: bool = False,
            created_at: Optional[datetime] = None,
            created_by: Optional[str] = None,
            confidence: float = DEFAULT_CONFIDENCE,
            manual_annotation: bool = DEFAULT_MANUAL_ANNOTATION,
            last_edited_at: Optional[datetime] = None,
            last_edited_by: Optional[str] = None,
        ) -> None:
            """Add a classification instance to the current frame.

            Args:
                classification_instance: The classification instance to add.
                overwrite: Whether to overwrite existing instance data (default is False).
                created_at: Optional creation timestamp.
                created_by: Optional creator identifier.
                confidence: Confidence score (default is DEFAULT_CONFIDENCE).
                manual_annotation: Flag indicating manual annotation (default is DEFAULT_MANUAL_ANNOTATION).
                last_edited_at: Optional last edited timestamp.
                last_edited_by: Optional last editor identifier.

            Raises:
                LabelRowError: If the classification instance is already assigned to a different label row.
            """
            if created_at is None:
                created_at = datetime.now()

            if last_edited_at is None:
                last_edited_at = datetime.now()

            label_row = classification_instance.is_assigned_to_label_row()
            if label_row and self._label_row != label_row:
                raise LabelRowError(
                    "This classification instance is already assigned to a different label row. It cannot be "
                    "added to multiple label rows at once."
                )

            classification_instance.set_for_frames(
                self._frame,
                overwrite=overwrite,
                created_at=created_at,
                created_by=created_by,
                confidence=confidence,
                manual_annotation=manual_annotation,
                last_edited_at=last_edited_at,
                last_edited_by=last_edited_by,
            )

            if not label_row:
                self._label_row.add_classification_instance(classification_instance)

        def get_object_instances(self, filter_ontology_object: Optional[Object] = None) -> List[ObjectInstance]:
            """Get object instances for the current frame.

            Args:
                filter_ontology_object: Optionally filter by a specific ontology object.

            Returns:
                List[ObjectInstance]: All `ObjectInstance`s that match the filter.
            """
            return self._label_row.get_object_instances(
                filter_ontology_object=filter_ontology_object, filter_frames=self._frame
            )

        def get_classification_instances(
            self, filter_ontology_classification: Optional[Classification] = None
        ) -> List[ClassificationInstance]:
            """Get classification instances for the current frame.

            Args:
                filter_ontology_classification: Optionally filter by a specific ontology classification.

            Returns:
                List[ClassificationInstance]: All `ClassificationInstance`s that match the filter.
            """
            return self._label_row.get_classification_instances(
                filter_ontology_classification=filter_ontology_classification, filter_frames=self._frame
            )

        def _frame_level_data(self) -> LabelRowV2.FrameLevelImageGroupData:
            return self._label_row_read_only_data.frame_level_data[self._frame]

        def __repr__(self):
            return f"FrameView(label_row={self._label_row}, frame={self._frame})"

    @dataclass(frozen=True)
    class FrameLevelImageGroupData:
        """Data related to a specific frame in an image group.

        Attributes:
            image_hash: Hash of the image.
            image_title: Title of the image.
            file_type: Type of the file (e.g., JPEG, PNG).
            frame_number: The number of the frame in the sequence.
            width: Width of the image.
            height: Height of the image.
            data_link: Optional link to additional data related to the image.
        """

        image_hash: str
        image_title: str
        file_type: str
        frame_number: int
        width: int
        height: int
        data_link: Optional[str] = None

    @dataclass(frozen=True)
    class LabelRowReadOnlyDataImagesDataEntry:
        """Read-only data entry for image data in a label row.

        Attributes:
            index: Index of the image.
            title: Title of the image.
            file_type: Type of the file (e.g., JPEG, PNG).
            height: Height of the image.
            width: Width of the image.
            image_hash: Hash of the image.
        """

        index: int
        title: str
        file_type: str
        height: int
        width: int
        image_hash: str

    @dataclass(frozen=True)
    class LabelRowReadOnlyData:
        """Read-only data for a label row.

        Attributes:
            label_hash: Optional hash for the label. `None` if not initialized for labeling.
            created_at: Optional creation timestamp. `None` if not initialized for labeling.
            last_edited_at: Optional last edited timestamp. `None` if not initialized for labeling.
            data_hash: Hash of the data.
            data_type: Type of the data (e.g., IMAGE, VIDEO).
            label_status: Status of the label.
            annotation_task_status: Optional status of the annotation task.
            workflow_graph_node: Optional workflow graph node related to the data.
            is_shadow_data: Boolean indicating if the data is shadow data.
            number_of_frames: Number of frames in the data.
            duration: Optional duration of the video or sequence.
            fps: Optional frames per second for video.
            dataset_hash: Hash of the dataset.
            dataset_title: Title of the dataset.
            data_title: Title of the data.
            audio_codec: Codec for audio data.
            audio_bit_depth: Bit depth for audio data.
            audio_sample_rate: Sample Rate for audio data.
            audio_num_channels: Number of channels for audio data.
            width: Optional width of the data.
            height: Optional height of the data.
            data_link: Optional link to additional data.
            task_uuid: Optional task uuid of the data.
            priority: Optional priority of the data.
            file_type: Optional file type of the data.
            client_metadata: Optional metadata provided by the client.
            images_data: Optional list of image data entries.
            branch_name: Name of the branch associated with the data.
            frame_level_data: Mapping from frame number to frame level data.
            image_hash_to_frame: Mapping from image hash to frame number.
            frame_to_image_hash: Mapping from frame number to image hash.
            is_valid: Boolean indicating if the data is valid.
            assigned_user_email: Optional email of the user assigned to the data.
            last_actioned_by_user_email: Optional email of the user who last actioned the data.
        """

        label_hash: Optional[str]
        created_at: Optional[datetime]
        last_edited_at: Optional[datetime]
        data_hash: str
        group_hash: Optional[str]
        data_type: DataType
        backing_item_uuid: Optional[UUID]
        label_status: LabelStatus
        annotation_task_status: Optional[AnnotationTaskStatus]
        workflow_graph_node: Optional[WorkflowGraphNode]
        is_shadow_data: bool
        number_of_frames: int
        duration: Optional[float]
        fps: Optional[float]
        dataset_hash: str
        dataset_title: str
        data_title: str
        audio_codec: Optional[str]
        audio_bit_depth: Optional[int]
        audio_num_channels: Optional[int]
        audio_sample_rate: Optional[int]
        width: Optional[int]
        height: Optional[int]
        data_link: Optional[str]
        task_uuid: Optional[UUID]
        priority: Optional[float]
        file_type: Optional[str]
        client_metadata: Optional[Dict[str, Any]]
        images_data: Optional[List[LabelRowV2.LabelRowReadOnlyDataImagesDataEntry]]
        branch_name: str
        frame_level_data: Dict[int, LabelRowV2.FrameLevelImageGroupData] = field(default_factory=dict)
        image_hash_to_frame: Dict[str, int] = field(default_factory=dict)
        frame_to_image_hash: Dict[int, str] = field(default_factory=dict)
        is_valid: bool = field(default=True)
        assigned_user_email: Optional[str] = None
        last_actioned_by_user_email: Optional[str] = None

    def _to_object_answers(self) -> Dict[str, Any]:
        ret: Dict[str, Any] = {}
        for obj in self._objects_map.values():
            all_static_answers = self._get_all_static_answers(obj)
            ret[obj.object_hash] = {
                "classifications": list(reversed(all_static_answers)),
                "objectHash": obj.object_hash,
            }

            # At some point, we also want to add these to the other modalities
            if not is_geometric(self.data_type):
                # For non-frame entities, all annotations exist only on one frame
                annotation = obj.get_annotation(0)
                object_answer_dict = ret[obj.object_hash]
                object_answer_dict["createdBy"] = annotation.created_by
                object_answer_dict["createdAt"] = format_datetime_to_long_string_optional(annotation.created_at)
                object_answer_dict["lastEditedBy"] = annotation.last_edited_by
                object_answer_dict["lastEditedAt"] = format_datetime_to_long_string_optional(annotation.last_edited_at)
                object_answer_dict["manualAnnotation"] = annotation.manual_annotation
                object_answer_dict["featureHash"] = obj.feature_hash
                object_answer_dict["name"] = obj.ontology_item.name
                object_answer_dict["color"] = obj.ontology_item.color
                object_answer_dict["shape"] = obj.ontology_item.shape.value
                object_answer_dict["value"] = _lower_snake_case(obj.ontology_item.name)

                if self.file_type == "text/html":
                    if obj.range_html is None:
                        raise LabelRowError("Html annotations should have range_html set within the TextCoordinates")
                    object_answer_dict["range_html"] = [x.to_dict() for x in obj.range_html]
                    object_answer_dict["range"] = []
                else:
                    if obj.range_list is None:
                        raise LabelRowError("Non-geometric annotations should have range set within the Coordinates")
                    object_answer_dict["range"] = [[range.start, range.end] for range in obj.range_list]

        return ret

    def _to_object_actions(self) -> Dict[str, Any]:
        ret: Dict[str, Any] = {}
        for obj in self._objects_map.values():
            all_static_answers = self._dynamic_answers_to_encord_dict(obj)
            if len(all_static_answers) == 0:
                continue
            ret[obj.object_hash] = {
                "actions": list(reversed(all_static_answers)),
                "objectHash": obj.object_hash,
            }
        return ret

    def _to_classification_answers(self) -> Dict[str, Any]:
        ret: Dict[str, Any] = {}
        for classification in self._classifications_map.values():
            all_static_answers = classification.get_all_static_answers()
            classifications = [answer.to_encord_dict() for answer in all_static_answers if answer.is_answered()]
            ret[classification.classification_hash] = {
                "classifications": list(reversed(classifications)),
                "classificationHash": classification.classification_hash,
                "featureHash": classification.feature_hash,
            }

            # At some point, we also want to add these to the other modalities
            if not is_geometric(self.data_type):
                annotation = classification.get_annotations()[0]

                # For non-geometric data, classifications apply to whole file
                ret[classification.classification_hash]["range"] = []
                ret[classification.classification_hash]["createdBy"] = annotation.created_by
                ret[classification.classification_hash]["createdAt"] = format_datetime_to_long_string_optional(
                    annotation.created_at
                )
                ret[classification.classification_hash]["lastEditedBy"] = annotation.last_edited_by
                ret[classification.classification_hash]["lastEditedAt"] = format_datetime_to_long_string_optional(
                    annotation.last_edited_at
                )
                ret[classification.classification_hash]["manualAnnotation"] = annotation.manual_annotation

        return ret

    @staticmethod
    def _get_all_static_answers(object_instance: ObjectInstance) -> List[Dict[str, Any]]:
        ret = []
        for answer in object_instance._get_all_static_answers():
            d_opt = answer.to_encord_dict()
            if d_opt is not None:
                ret.append(d_opt)
        return ret

    @staticmethod
    def _dynamic_answers_to_encord_dict(object_instance: ObjectInstance) -> List[Dict[str, Any]]:
        ret = []
        for answer, ranges in object_instance._get_all_dynamic_answers():
            d_opt = answer.to_encord_dict(ranges)
            if d_opt is not None:
                ret.append(d_opt)
        return ret

    def _to_encord_data_units(self) -> Dict[str, Any]:
        frame_level_data = self._label_row_read_only_data.frame_level_data
        return {value.image_hash: self._to_encord_data_unit(value) for value in frame_level_data.values()}

    def _to_encord_data_unit(self, frame_level_data: FrameLevelImageGroupData) -> Dict[str, Any]:
        ret: Dict[str, Any] = {}

        data_type = self._label_row_read_only_data.data_type

        if data_type == DataType.IMG_GROUP:
            data_sequence: Union[str, int] = str(frame_level_data.frame_number)

        elif (
            data_type == DataType.VIDEO
            or data_type == DataType.DICOM
            or data_type == DataType.IMAGE
            or data_type == DataType.NIFTI
            or data_type == DataType.PDF
            or data_type == DataType.SCENE
        ):
            data_sequence = frame_level_data.frame_number

        elif data_type == DataType.AUDIO or data_type == DataType.PLAIN_TEXT:
            data_sequence = 0

        elif data_type == DataType.DICOM_STUDY:
            pass
        elif data_type == DataType.GROUP:
            data_sequence = 0

        elif data_type == DataType.MISSING_DATA_TYPE:
            raise NotImplementedError(f"The data type {data_type} is not implemented yet.")

        else:
            exhaustive_guard(data_type)

        ret["data_hash"] = frame_level_data.image_hash
        ret["data_title"] = frame_level_data.image_title

        if data_type != DataType.DICOM and data_type != DataType.GROUP:
            ret["data_link"] = frame_level_data.data_link

        ret["data_type"] = frame_level_data.file_type
        ret["data_sequence"] = data_sequence

        if self.data_type == DataType.AUDIO:
            ret["audio_codec"] = self._label_row_read_only_data.audio_codec
            ret["audio_sample_rate"] = self._label_row_read_only_data.audio_sample_rate
            ret["audio_bit_depth"] = self._label_row_read_only_data.audio_bit_depth
            ret["audio_num_channels"] = self._label_row_read_only_data.audio_num_channels
        elif self.data_type == DataType.PLAIN_TEXT:
            pass
        elif self.data_type == DataType.GROUP:
            ret["width"] = 0
            ret["height"] = 0
            ret["data_link"] = ""
        elif self.data_type == DataType.SCENE:
            ret["width"] = 0
            ret["height"] = 0
            ret["data_link"] = ""
            ret["data_fps"] = 0
        elif self.data_type == DataType.PDF:
            ret["height"] = 0
            ret["width"] = 0
            ret["data_fps"] = 0
        elif (
            self.data_type == DataType.IMAGE
            or self.data_type == DataType.NIFTI
            or self.data_type == DataType.VIDEO
            or self.data_type == DataType.IMG_GROUP
            or self.data_type == DataType.DICOM
            or self.data_type == DataType.DICOM_STUDY
        ):
            ret["width"] = frame_level_data.width
            ret["height"] = frame_level_data.height
        elif self.data_type == DataType.MISSING_DATA_TYPE:
            raise LabelRowError("Label row is missing data type.")
        else:
            exhaustive_guard(self.data_type)

        ret["labels"] = self._to_encord_labels(frame_level_data)

        if self._label_row_read_only_data.duration is not None and self.data_type != DataType.PLAIN_TEXT:
            ret["data_duration"] = self._label_row_read_only_data.duration

        if (
            self._label_row_read_only_data.fps is not None
            and self.data_type != DataType.AUDIO
            and self.data_type != DataType.PLAIN_TEXT
        ):
            ret["data_fps"] = self._label_row_read_only_data.fps

        return ret

    def _to_encord_labels(self, frame_level_data: FrameLevelImageGroupData) -> Dict[str, Any]:
        ret: Dict[str, Any] = {}
        data_type = self._label_row_read_only_data.data_type

        if data_type == DataType.IMAGE or data_type == DataType.IMG_GROUP or data_type == DataType.GROUP:
            # single frame
            frame = frame_level_data.frame_number
            ret.update(self._to_encord_label(frame))

        elif (
            # multi-frame
            data_type == DataType.VIDEO
            or data_type == DataType.DICOM
            or data_type == DataType.NIFTI
            or data_type == DataType.PDF
            or data_type == DataType.SCENE
        ):
            for frame in self._frame_to_hashes.keys():
                ret[str(frame)] = self._to_encord_label(frame)

        elif data_type == DataType.AUDIO or data_type == DataType.PLAIN_TEXT:
            return {}

        elif data_type == DataType.DICOM_STUDY:
            pass

        elif data_type == DataType.MISSING_DATA_TYPE:
            raise NotImplementedError(f"The data type {data_type} is not implemented yet.")

        else:
            exhaustive_guard(data_type)

        return ret

    def _to_encord_label(self, frame: int) -> Dict[str, Any]:
        ret: Dict[str, Any] = {}

        ret["objects"] = self._to_encord_objects_list(frame)
        ret["classifications"] = self._to_encord_classifications_list(frame)

        return ret

    def _to_encord_objects_list(self, frame: int) -> list:
        # Get objects for frame
        ret: List[dict] = []

        objects = self.get_object_instances(filter_frames=frame)
        for object_ in objects:
            encord_object = self._to_encord_object(object_, frame)
            ret.append(encord_object)
        return ret

    def _to_encord_object(
        self,
        object_: ObjectInstance,
        frame: int,
    ) -> Dict[str, Any]:
        ret: Dict[str, Any] = {}

        object_instance_annotation = object_.get_annotation(frame)
        coordinates = object_instance_annotation.coordinates
        ontology_hash = object_.ontology_item.feature_node_hash
        ontology_object = self._ontology.structure.get_child_by_hash(ontology_hash, type_=Object)

        ret["name"] = ontology_object.name
        ret["color"] = ontology_object.color
        ret["shape"] = ontology_object.shape.value
        ret["value"] = _lower_snake_case(ontology_object.name)
        ret["createdAt"] = format_datetime_to_long_string_optional(object_instance_annotation.created_at)
        ret["createdBy"] = object_instance_annotation.created_by
        ret["confidence"] = object_instance_annotation.confidence
        ret["objectHash"] = object_.object_hash
        ret["featureHash"] = ontology_object.feature_node_hash
        ret["manualAnnotation"] = object_instance_annotation.manual_annotation

        if object_instance_annotation.last_edited_at is not None:
            ret["lastEditedAt"] = format_datetime_to_long_string_optional(object_instance_annotation.last_edited_at)
        if object_instance_annotation.last_edited_by is not None:
            ret["lastEditedBy"] = object_instance_annotation.last_edited_by
        if object_instance_annotation.is_deleted is not None:
            ret["isDeleted"] = object_instance_annotation.is_deleted

        self._add_coordinates_to_encord_object(coordinates, frame, ret)

        return ret

    def _add_coordinates_to_encord_object(
        self, coordinates: Coordinates, frame: int, encord_object: Dict[str, Any]
    ) -> None:
        if isinstance(coordinates, BoundingBoxCoordinates):
            encord_object["boundingBox"] = coordinates.to_dict()
        elif isinstance(coordinates, RotatableBoundingBoxCoordinates):
            encord_object["rotatableBoundingBox"] = coordinates.to_dict()
        elif isinstance(coordinates, PolygonCoordinates):
            encord_object["polygon"] = coordinates.to_dict()
            encord_object["polygons"] = coordinates.to_dict(PolygonCoordsToDict.multiple_polygons)
        elif isinstance(coordinates, PolylineCoordinates):
            encord_object["polyline"] = coordinates.to_dict()
        elif isinstance(coordinates, PointCoordinate):
            encord_object["point"] = coordinates.to_dict()
        elif isinstance(coordinates, BitmaskCoordinates):
            frame_view = self.get_frame_view(frame)
            if not (
                frame_view.height == coordinates._encoded_bitmask.height
                and frame_view.width == coordinates._encoded_bitmask.width
            ):
                raise ValueError("Bitmask dimensions don't match the media dimensions")
            encord_object["bitmask"] = coordinates.to_dict()
        elif isinstance(coordinates, SkeletonCoordinates):
            encord_object["skeleton"] = coordinates.to_dict()
        elif isinstance(coordinates, CuboidCoordinates):
            encord_object["cuboid"] = coordinates.to_dict()
        else:
            raise NotImplementedError(f"adding coordinatees for this type not yet implemented {type(coordinates)}")

    def _to_encord_classifications_list(self, frame: int) -> list:
        ret: List[Dict[str, Any]] = []

        classifications = self.get_classification_instances(filter_frames=frame)
        for classification in classifications:
            encord_classification = self._to_encord_classification(classification, frame)
            ret.append(encord_classification)

        return ret

    def _to_encord_classification(self, classification: ClassificationInstance, frame: int) -> Dict[str, Any]:
        ret: Dict[str, Any] = {}

        annotation = classification.get_annotation(frame)
        classification_feature_hash = classification.ontology_item.feature_node_hash
        ontology_classification = self._ontology.structure.get_child_by_hash(
            classification_feature_hash, type_=Classification
        )
        attribute_hash = classification.ontology_item.attributes[0].feature_node_hash
        ontology_attribute = self._ontology.structure.get_child_by_hash(attribute_hash, type_=Attribute)

        ret["name"] = ontology_attribute.name
        ret["value"] = _lower_snake_case(ontology_attribute.name)
        ret["createdAt"] = format_datetime_to_long_string_optional(annotation.created_at)
        ret["createdBy"] = annotation.created_by
        ret["confidence"] = annotation.confidence
        ret["featureHash"] = ontology_classification.feature_node_hash
        ret["classificationHash"] = classification.classification_hash
        ret["manualAnnotation"] = annotation.manual_annotation

        if annotation.last_edited_at is not None:
            ret["lastEditedAt"] = format_datetime_to_long_string_optional(annotation.last_edited_at)
        if annotation.last_edited_by is not None:
            ret["lastEditedBy"] = annotation.last_edited_by

        return ret

    def _is_classification_already_present_on_frames(
        self, classification: Classification, frames: Iterable[int]
    ) -> Set[int]:
        present_frames = self._classifications_to_frames.get(classification, set())

        return present_frames.intersection(frames)

    def _is_classification_already_present_on_ranges(
        self,
        classification: Classification,
        ranges: Ranges,
    ) -> Ranges:
        range_manager = self._classifications_to_ranges.get(classification, RangeManager())
        return range_manager.intersection(ranges)

    def _add_frames_to_classification(self, classification: Classification, frames: Iterable[int]) -> None:
        self._classifications_to_frames[classification].update(frames)

    def _add_ranges_to_classification(self, classification: Classification, ranges_to_add: Ranges) -> None:
        self._classifications_to_ranges[classification].add_ranges(ranges_to_add)

    def _remove_frames_from_classification(self, classification: Classification, frames: Iterable[int]) -> None:
        present_frames = self._classifications_to_frames.get(classification, set())
        for frame in frames:
            present_frames.remove(frame)

    def _remove_ranges_from_classification(self, classification: Classification, ranges_to_remove: Ranges) -> None:
        self._classifications_to_ranges[classification].remove_ranges(ranges_to_remove)

    def _add_to_frame_to_hashes_map(
        self, label_item: Union[ObjectInstance, ClassificationInstance], frames: Iterable[int]
    ) -> None:
        for frame in frames:
            self.add_to_single_frame_to_hashes_map(label_item, frame)

    def _remove_from_frame_to_hashes_map(self, frames: Iterable[int], item_hash: str):
        for frame in frames:
            self._frame_to_hashes[frame].remove(item_hash)

    def _parse_label_row_metadata(self, label_row_metadata: LabelRowMetadata) -> LabelRowV2.LabelRowReadOnlyData:
        data_type = DataType.from_upper_case_string(label_row_metadata.data_type)

        return LabelRowV2.LabelRowReadOnlyData(
            label_hash=label_row_metadata.label_hash,
            branch_name=label_row_metadata.branch_name,
            data_hash=label_row_metadata.data_hash,
            group_hash=label_row_metadata.group_hash,
            data_title=label_row_metadata.data_title,
            dataset_hash=label_row_metadata.dataset_hash,
            dataset_title=label_row_metadata.dataset_title,
            data_type=data_type,
            data_link=label_row_metadata.data_link,
            label_status=label_row_metadata.label_status,
            annotation_task_status=label_row_metadata.annotation_task_status,
            workflow_graph_node=label_row_metadata.workflow_graph_node,
            is_shadow_data=label_row_metadata.is_shadow_data,
            created_at=label_row_metadata.created_at,
            last_edited_at=label_row_metadata.last_edited_at,
            duration=label_row_metadata.duration,
            fps=label_row_metadata.frames_per_second,
            number_of_frames=label_row_metadata.number_of_frames,
            audio_codec=label_row_metadata.audio_codec,
            audio_sample_rate=label_row_metadata.audio_sample_rate,
            audio_num_channels=label_row_metadata.audio_num_channels,
            audio_bit_depth=label_row_metadata.audio_bit_depth,
            width=label_row_metadata.width,
            height=label_row_metadata.height,
            task_uuid=label_row_metadata.task_uuid,
            priority=label_row_metadata.priority,
            images_data=None
            if label_row_metadata.images_data is None
            else [LabelRowV2.LabelRowReadOnlyDataImagesDataEntry(**data) for data in label_row_metadata.images_data],
            client_metadata=label_row_metadata.client_metadata,
            file_type=label_row_metadata.file_type,
            is_valid=label_row_metadata.is_valid,
            backing_item_uuid=label_row_metadata.backing_item_uuid,
            assigned_user_email=label_row_metadata.assigned_user_email,
            last_actioned_by_user_email=label_row_metadata.last_actioned_by_user_email,
        )

    def _parse_label_row_dict(self, label_row_dict: dict) -> LabelRowReadOnlyData:
        frame_level_data = self._parse_image_group_frame_level_data(label_row_dict["data_units"])
        image_hash_to_frame = {item.image_hash: item.frame_number for item in frame_level_data.values()}
        frame_to_image_hash = {item.frame_number: item.image_hash for item in frame_level_data.values()}
        data_type = DataType(label_row_dict["data_type"])

        audio_codec = None
        audio_sample_rate = None
        audio_num_channels = None
        audio_bit_depth = None

        if data_type == DataType.VIDEO or data_type == DataType.IMAGE:
            data_dict = list(label_row_dict["data_units"].values())[0]
            data_link = data_dict["data_link"]
            file_type = data_dict.get("data_type")
            # Dimensions should be always there
            # But we have some older entries that don't have them
            # So setting them to None for now until the format is not guaranteed to be enforced
            height = data_dict.get("height")
            width = data_dict.get("width")

        elif data_type == DataType.GROUP or data_type == DataType.SCENE:
            data_dict = list(label_row_dict["data_units"].values())[0]
            file_type = data_dict.get("data_type")
            data_link = None
            width = None
            height = None

        elif data_type == DataType.DICOM or data_type == DataType.NIFTI:
            dicom_dict = list(label_row_dict["data_units"].values())[0]
            data_link = None
            file_type = dicom_dict["data_type"]

            height = dicom_dict["height"]
            width = dicom_dict["width"]

        elif data_type == DataType.AUDIO:
            data_dict = list(label_row_dict["data_units"].values())[0]
            data_link = data_dict["data_link"]
            file_type = data_dict.get("data_type")
            height = None
            width = None
            audio_codec = data_dict["audio_codec"]
            audio_sample_rate = data_dict["audio_sample_rate"]
            audio_num_channels = data_dict["audio_num_channels"]
            audio_bit_depth = data_dict["audio_bit_depth"]

        elif data_type == DataType.PLAIN_TEXT:
            data_dict = list(label_row_dict["data_units"].values())[0]
            data_link = data_dict["data_link"]
            file_type = data_dict.get("data_type")
            height = None
            width = None

        elif data_type == DataType.IMG_GROUP:
            data_link = None
            file_type = None
            height = None
            width = None

        elif data_type == DataType.DICOM_STUDY:
            pass

        elif data_type == DataType.PLAIN_TEXT or data_type == DataType.PDF:
            data_dict = list(label_row_dict["data_units"].values())[0]
            data_link = data_dict["data_link"]
            file_type = data_dict.get("data_type")
            height = None
            width = None

        elif data_type == DataType.MISSING_DATA_TYPE:
            raise NotImplementedError(f"The data type {data_type} is not implemented yet.")

        else:
            exhaustive_guard(data_type)

        return LabelRowV2.LabelRowReadOnlyData(
            label_hash=label_row_dict["label_hash"],
            branch_name=label_row_dict["branch_name"],
            dataset_hash=label_row_dict["dataset_hash"],
            dataset_title=label_row_dict["dataset_title"],
            data_title=label_row_dict["data_title"],
            data_hash=label_row_dict["data_hash"],
            group_hash=label_row_dict.get("group_hash", self._label_row_read_only_data.group_hash),
            data_type=data_type,
            label_status=LabelStatus(label_row_dict["label_status"]),
            annotation_task_status=label_row_dict.get("annotation_task_status"),
            workflow_graph_node=label_row_dict.get(
                "workflow_graph_node", self._label_row_read_only_data.workflow_graph_node
            ),
            is_shadow_data=self.is_shadow_data,
            created_at=parse_datetime_optional(label_row_dict["created_at"]),
            last_edited_at=parse_datetime_optional(label_row_dict["last_edited_at"]),
            frame_level_data=frame_level_data,
            image_hash_to_frame=image_hash_to_frame,
            frame_to_image_hash=frame_to_image_hash,
            duration=self.duration,
            fps=self.fps,
            number_of_frames=self.number_of_frames,
            data_link=data_link,
            height=height,
            width=width,
            audio_codec=audio_codec,
            audio_sample_rate=audio_sample_rate,
            audio_num_channels=audio_num_channels,
            audio_bit_depth=audio_bit_depth,
            task_uuid=label_row_dict.get("task_uuid", self._label_row_read_only_data.task_uuid),
            priority=label_row_dict.get("priority", self._label_row_read_only_data.priority),
            client_metadata=label_row_dict.get("client_metadata", self._label_row_read_only_data.client_metadata),
            images_data=label_row_dict.get("images_data", self._label_row_read_only_data.images_data),
            file_type=file_type,
            is_valid=bool(label_row_dict.get("is_valid", True)),
            backing_item_uuid=self.backing_item_uuid,
            assigned_user_email=label_row_dict.get(
                "assigned_user_email", self._label_row_read_only_data.assigned_user_email
            ),
            last_actioned_by_user_email=label_row_dict.get(
                "last_actioned_by_user_email", self._label_row_read_only_data.last_actioned_by_user_email
            ),
        )

    def _parse_labels_from_dict(self, label_row_dict: dict):
        classification_answers = label_row_dict["classification_answers"]
        object_answers = label_row_dict["object_answers"]

        for data_unit in label_row_dict["data_units"].values():
            data_type = DataType(label_row_dict["data_type"])

            if data_type == DataType.IMG_GROUP or data_type == DataType.IMAGE:
                frame = int(data_unit["data_sequence"])
                self._add_object_instances_from_objects(data_unit["labels"].get("objects", []), frame)
                self._add_classification_instances_from_classifications(
                    data_unit["labels"].get("classifications", []),
                    classification_answers,
                    frame,
                )
                self._add_objects_answers(object_answers)

            elif (
                data_type == DataType.VIDEO
                or data_type == DataType.DICOM
                or data_type == DataType.NIFTI
                or data_type == DataType.PDF
                or data_type == DataType.SCENE
            ):
                for frame, frame_data in data_unit["labels"].items():
                    frame_num = int(frame)
                    self._add_object_instances_from_objects(frame_data["objects"], frame_num)
                    self._add_classification_instances_from_classifications(
                        frame_data["classifications"], classification_answers, frame_num
                    )
                    self._add_frame_metadata(frame_num, frame_data.get("metadata"))
                self._add_objects_answers(object_answers)

            elif data_type == DataType.GROUP:
                for frame, frame_data in data_unit["labels"].items():
                    frame_num = int(frame)
                    self._add_classification_instances_from_classifications(
                        frame_data["classifications"], classification_answers, frame_num
                    )
                    self._add_frame_metadata(frame_num, frame_data.get("metadata"))

            elif data_type == DataType.DICOM_STUDY:
                pass  # TODO: _add_object_answers a NO-OP here as well?

            elif data_type == DataType.MISSING_DATA_TYPE:
                raise NotImplementedError(f"Got an unexpected data type `{data_type}`")

            elif data_type == DataType.AUDIO or data_type == DataType.PLAIN_TEXT:
                is_html = data_unit["data_type"] == "text/html"
                self._add_objects_instances_from_objects_without_frames(object_answers, html=is_html)
                self._add_classification_instances_from_classifications_without_frames(classification_answers)
            else:
                exhaustive_guard(data_type)

            self._add_data_unit_metadata(data_type, data_unit.get("metadata"))

        self._add_action_answers(label_row_dict)

    def _add_frame_metadata(self, frame: int, metadata: Optional[Dict[str, str]]):
        if metadata is not None:
            self._frame_metadata[frame] = DICOMSliceMetadata.from_dict(metadata)
        else:
            self._frame_metadata[frame] = None

    def _add_data_unit_metadata(self, data_type: DataType, metadata: Optional[Dict[str, str]]) -> None:
        if metadata is None:
            self._metadata = None
            return

        if data_type == DataType.DICOM:
            self._metadata = DICOMSeriesMetadata.from_dict(metadata)
        elif data_type == DataType.GROUP:
            self._metadata = DataGroupMetadata.from_dict(metadata)
        else:
            log.warning(
                f"Unexpected metadata for the data type: {data_type}. Please update the Encord SDK to the latest version."
            )

    def _add_objects_instances_from_objects_without_frames(
        self,
        object_answers: dict,
        html: bool = False,
    ):
        if html:
            for object_answer in object_answers.values():
                object_instance = self._create_new_html_object_instance(object_answer, object_answer["range_html"])
                self.add_object_instance(object_instance)

        else:
            for object_answer in object_answers.values():
                ranges: Ranges = []
                for range_elem in object_answer["range"]:
                    ranges.append(Range(range_elem[0], range_elem[1]))

                object_instance = self._create_new_object_instance_with_ranges(object_answer, ranges)
                self.add_object_instance(object_instance)

    def _add_object_instances_from_objects(
        self,
        objects_list: List[dict],
        frame: int,
    ) -> None:
        for frame_object_label in objects_list:
            object_hash = frame_object_label["objectHash"]
            if object_hash not in self._objects_map:
                object_instance = self._create_new_object_instance(frame_object_label, frame)
                self.add_object_instance(object_instance)
            else:
                self._add_coordinates_to_object_instance(frame_object_label, frame)

    def _add_objects_answers(self, object_answers: dict):
        for answer in object_answers.values():
            object_hash = answer["objectHash"]
            # In cases when we had an object, added some attributes for this object, and then removed the object,
            # in some label rows we still have such "orphaned" answers.
            # To avoid parser errors, we're omitting attributes for the object that is not in label rows.
            if object_instance := self._objects_map.get(object_hash):
                answer_list = answer["classifications"]
                object_instance.set_answer_from_list(answer_list)

    def _add_action_answers(self, label_row_dict: dict):
        for answer in label_row_dict["object_actions"].values():
            object_hash = answer["objectHash"]
            object_instance = self._objects_map[object_hash]

            answer_list = answer["actions"]
            object_instance.set_answer_from_list(answer_list)

    def _create_new_object_instance(self, frame_object_label: dict, frame: int) -> ObjectInstance:
        ontology = self._ontology.structure
        feature_hash = frame_object_label["featureHash"]
        object_hash = frame_object_label["objectHash"]

        label_class = ontology.get_child_by_hash(feature_hash, type_=Object)
        object_instance = ObjectInstance(label_class, object_hash=object_hash)

        coordinates = self._get_coordinates(frame_object_label)
        object_frame_instance_info = ObjectInstance.FrameInfo.from_dict(frame_object_label)

        object_instance.set_for_frames(
            coordinates=coordinates,
            frames=frame,
            created_at=object_frame_instance_info.created_at,
            created_by=object_frame_instance_info.created_by,
            last_edited_at=object_frame_instance_info.last_edited_at,
            last_edited_by=object_frame_instance_info.last_edited_by,
            confidence=object_frame_instance_info.confidence,
            manual_annotation=object_frame_instance_info.manual_annotation,
            reviews=object_frame_instance_info.reviews,
            is_deleted=object_frame_instance_info.is_deleted,
        )
        return object_instance

    # This is only to be used by non-frame modalities (e.g. Audio)
    def _create_new_object_instance_with_ranges(
        self,
        object_answer: dict,
        ranges: Ranges,
    ) -> ObjectInstance:
        feature_hash = object_answer["featureHash"]
        object_hash = object_answer["objectHash"]

        label_class = self._ontology.structure.get_child_by_hash(feature_hash, type_=Object)

        frame_info_dict = {k: v for k, v in object_answer.items() if v is not None}
        frame_info_dict.setdefault("confidence", 1.0)  # confidence sometimes not present.
        object_frame_instance_info = ObjectInstance.FrameInfo.from_dict(frame_info_dict)

        expected_shape: Shape
        coordinates: Union[AudioCoordinates, TextCoordinates]
        if self._label_row_read_only_data.data_type == DataType.AUDIO:
            expected_shape = Shape.AUDIO
            coordinates = AudioCoordinates(range=ranges)
        elif self._label_row_read_only_data.data_type == DataType.PLAIN_TEXT:
            expected_shape = Shape.TEXT
            coordinates = TextCoordinates(range=ranges)
        else:
            unknown_data_type = self._label_row_read_only_data.data_type
            raise RuntimeError(f"Unexpected data type[{unknown_data_type}] for range based objects")
        if label_class.shape != expected_shape:
            raise LabelRowError("Unsupported object shape for data type")
        object_instance = ObjectInstance(label_class, object_hash=object_hash)

        object_instance.set_for_frames(
            coordinates,
            frames=0,
            created_at=object_frame_instance_info.created_at,
            created_by=object_frame_instance_info.created_by,
            confidence=object_frame_instance_info.confidence,
            manual_annotation=object_frame_instance_info.manual_annotation,
            last_edited_at=object_frame_instance_info.last_edited_at,
            last_edited_by=object_frame_instance_info.last_edited_by,
            reviews=object_frame_instance_info.reviews,
            overwrite=True,
            # Always overwrite during label row dict parsing, as older dicts known to have duplicates
        )
        answer_list = object_answer["classifications"]
        object_instance.set_answer_from_list(answer_list)

        return object_instance

    def _create_new_html_object_instance(
        self,
        object_answer: dict,
        range_html: dict,
    ) -> ObjectInstance:
        feature_hash = object_answer["featureHash"]
        object_hash = object_answer["objectHash"]

        label_class = self._ontology.structure.get_child_by_hash(feature_hash, type_=Object)

        frame_info_dict = {k: v for k, v in object_answer.items() if v is not None}
        frame_info_dict.setdefault("confidence", 1.0)  # confidence sometimes not present.
        object_frame_instance_info = ObjectInstance.FrameInfo.from_dict(frame_info_dict)

        object_instance = ObjectInstance(label_class, object_hash=object_hash)
        object_instance.set_for_frames(
            HtmlCoordinates(range=[HtmlRange.from_dict(x) for x in range_html]),
            frames=0,
            created_at=object_frame_instance_info.created_at,
            created_by=object_frame_instance_info.created_by,
            confidence=object_frame_instance_info.confidence,
            manual_annotation=object_frame_instance_info.manual_annotation,
            last_edited_at=object_frame_instance_info.last_edited_at,
            last_edited_by=object_frame_instance_info.last_edited_by,
            reviews=object_frame_instance_info.reviews,
            overwrite=True,
            # Always overwrite during label row dict parsing, as older dicts known to have duplicates
        )
        answer_list = object_answer["classifications"]
        object_instance.set_answer_from_list(answer_list)

        return object_instance

    def _add_coordinates_to_object_instance(
        self,
        frame_object_label: dict,
        frame: int = 0,
    ) -> None:
        object_hash = frame_object_label["objectHash"]
        object_instance = self._objects_map[object_hash]

        coordinates = self._get_coordinates(frame_object_label)
        object_frame_instance_info = ObjectInstance.FrameInfo.from_dict(frame_object_label)

        object_instance.set_for_frames(
            coordinates=coordinates,
            frames=frame,
            created_at=object_frame_instance_info.created_at,
            created_by=object_frame_instance_info.created_by,
            last_edited_at=object_frame_instance_info.last_edited_at,
            last_edited_by=object_frame_instance_info.last_edited_by,
            confidence=object_frame_instance_info.confidence,
            manual_annotation=object_frame_instance_info.manual_annotation,
            reviews=object_frame_instance_info.reviews,
            is_deleted=object_frame_instance_info.is_deleted,
        )

    def _get_coordinates(self, frame_object_label: dict) -> Coordinates:
        if "boundingBox" in frame_object_label:
            return BoundingBoxCoordinates.from_dict(frame_object_label)
        if "rotatableBoundingBox" in frame_object_label:
            return RotatableBoundingBoxCoordinates.from_dict(frame_object_label)
        elif "polygon" in frame_object_label or "polygons" in frame_object_label:
            return PolygonCoordinates.from_dict(frame_object_label)
        elif "point" in frame_object_label:
            return PointCoordinate.from_dict(frame_object_label)
        elif "polyline" in frame_object_label:
            return PolylineCoordinates.from_dict(frame_object_label)
        elif "skeleton" in frame_object_label:

            def _with_visibility_enum(point: dict):
                if point.get(Visibility.INVISIBLE.value):
                    point["visibility"] = Visibility.INVISIBLE
                elif point.get(Visibility.OCCLUDED.value):
                    point["visibility"] = Visibility.OCCLUDED
                elif point.get(Visibility.VISIBLE.value):
                    point["visibility"] = Visibility.VISIBLE
                return point

            values = [_with_visibility_enum(pnt) for pnt in frame_object_label["skeleton"].values()]
            skeleton_frame_object_label = {
                "name": frame_object_label["name"],
                "values": values,
            }
            return SkeletonCoordinates.from_dict(skeleton_frame_object_label)
        elif "bitmask" in frame_object_label:
            return BitmaskCoordinates.from_dict(frame_object_label)
        elif "cuboid" in frame_object_label:
            return CuboidCoordinates.from_dict(frame_object_label)
        else:
            raise NotImplementedError(f"Getting coordinates for `{frame_object_label}` is not supported yet.")

    def _add_classification_instances_from_classifications(
        self, classifications_list: List[dict], classification_answers: dict, frame: int
    ):
        for frame_classification_label in classifications_list:
            classification_hash = frame_classification_label["classificationHash"]
            if classification_hash in self._classifications_map:
                self._add_frames_to_classification_instance(frame_classification_label, frame)
            elif classification_instance := self._create_new_classification_instance(
                frame_classification_label, frame, classification_answers
            ):
                self.add_classification_instance(classification_instance)

    def _add_classification_instances_from_classifications_without_frames(
        self,
        classification_answers: dict,
    ):
        for classification_answer in classification_answers.values():
            classification_instance = self._create_new_classification_instance_with_ranges(classification_answer)
            self.add_classification_instance(classification_instance)

    def _parse_image_group_frame_level_data(self, label_row_data_units: dict) -> Dict[int, FrameLevelImageGroupData]:
        frame_level_data: Dict[int, LabelRowV2.FrameLevelImageGroupData] = {}
        for payload in label_row_data_units.values():
            frame_number = int(payload["data_sequence"])
            frame_level_image_group_data = self.FrameLevelImageGroupData(
                image_hash=payload["data_hash"],
                image_title=payload["data_title"],
                data_link=payload.get("data_link"),
                file_type=payload["data_type"],
                frame_number=frame_number,
                # Dimensions should be always there
                # But we have some older entries that don't have them
                # So setting them to 0 for now until the format is not guaranteed to be enforced
                width=payload.get("width", 0),
                height=payload.get("height", 0),
            )
            frame_level_data[frame_number] = frame_level_image_group_data
        return frame_level_data

    def _create_new_classification_instance(
        self,
        frame_classification_label: dict,
        frame: int,
        classification_answers: dict,
    ) -> Optional[ClassificationInstance]:
        feature_hash = frame_classification_label["featureHash"]
        classification_hash = frame_classification_label["classificationHash"]

        label_class = self._ontology.structure.get_child_by_hash(feature_hash, type_=Classification)
        classification_instance = ClassificationInstance(label_class, classification_hash=classification_hash)

        frame_view = ClassificationInstance.FrameData.from_dict(frame_classification_label)
        classification_instance.set_for_frames(
            frame,
            created_at=frame_view.created_at,
            created_by=frame_view.created_by,
            confidence=frame_view.confidence,
            manual_annotation=frame_view.manual_annotation,
            last_edited_at=frame_view.last_edited_at,
            last_edited_by=frame_view.last_edited_by,
            reviews=frame_view.reviews,
            overwrite=True,  # Always overwrite during label row dict parsing, as older dicts known to have duplicates
        )

        # For some older label rows we might have a classification entry, but without an assigned answer.
        # These cases are equivalent to not having classifications at all, so just ignoring them
        if classification_answer := classification_answers.get(classification_hash):
            answers_dict = classification_answer["classifications"]
            self._add_static_answers_from_dict(classification_instance, answers_dict)
            return classification_instance

        return None

    # This is only to be used by non-frame modalities (e.g. Audio)
    def _create_new_classification_instance_with_ranges(self, classification_answer: dict) -> ClassificationInstance:
        feature_hash = classification_answer["featureHash"]
        classification_hash = classification_answer["classificationHash"]

        label_class = self._ontology.structure.get_child_by_hash(feature_hash, type_=Classification)

        range_view = ClassificationInstance.FrameData.from_dict(classification_answer)

        classification_instance = ClassificationInstance(
            label_class, classification_hash=classification_hash, range_only=True
        )

        # For non-geometric data, the classification will always be treated as being on frame=0,
        # which is the entire file
        classification_instance.set_for_frames(
            created_at=range_view.created_at,
            created_by=range_view.created_by,
            confidence=range_view.confidence,
            manual_annotation=range_view.manual_annotation,
            last_edited_at=range_view.last_edited_at,
            last_edited_by=range_view.last_edited_by,
            reviews=range_view.reviews,
            overwrite=True,  # Always overwrite during label row dict parsing, as older dicts known to have duplicates
        )
        answers_dict = classification_answer["classifications"]
        self._add_static_answers_from_dict(classification_instance, answers_dict)

        return classification_instance

    def _add_static_answers_from_dict(
        self, classification_instance: ClassificationInstance, answers_list: List[dict]
    ) -> None:
        classification_instance.set_answer_from_list(answers_list)

    def _add_frames_to_classification_instance(self, frame_classification_label: dict, frame: int) -> None:
        object_hash = frame_classification_label["classificationHash"]
        classification_instance = self._classifications_map[object_hash]
        frame_view = ClassificationInstance.FrameData.from_dict(frame_classification_label)
        classification_instance.set_frame_data(frame_view, frame)

    def _check_labelling_is_initalised(self):
        if not self.is_labelling_initialised:
            raise LabelRowError(
                "For this operation you will need to initialise labelling first. Call the `.initialise_labels()` "
                "to do so first."
            )

    def _method_not_supported_for_audio(self, range_only: bool = False):
        if self.data_type == DataType.AUDIO and not range_only:
            raise LabelRowError("This method is not supported for audio.")

    def __repr__(self) -> str:
        return f"LabelRowV2(label_hash={self.label_hash}, data_hash={self.data_hash}, data_title={self.data_title})"


def _frame_views_to_frame_numbers(
    frame_views: Sequence[Union[ObjectInstance.Annotation, ClassificationInstance.Annotation, LabelRowV2.FrameView]],
) -> List[int]:
    return [frame_view.frame for frame_view in frame_views]
